#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import os
import unittest

import mesh.log as log
from mesh.context import Mesh
from mesh.macro import ServiceLoader
from mesh.prsim import Transport, Routable, Metadata


class TestServiceLoad(unittest.TestCase):

    def test_transport_split(self):
        os.environ.setdefault("mesh.address", "127.0.0.1:17304")

        transport = Routable.of(ServiceLoader.load(Transport).get("mesh"))

        session = transport.with_address("127.0.0.1:7304").local().open('session_id_001', {
            Metadata.MESH_VERSION.key(): '',
            Metadata.MESH_TECH_PROVIDER_CODE.key(): 'LX',
            Metadata.MESH_TRACE_ID.key(): Mesh.context().get_trace_id(),
            Metadata.MESH_TOKEN.key(): 'x',
            Metadata.MESH_SESSION_ID.key(): 'session_id_001',
            Metadata.MESH_TARGET_NODE_ID.key(): 'LX0000000000000',
        })
        session2 = transport.with_address("127.0.0.1:7304").local().open('session_id_001', {
            Metadata.MESH_VERSION.key(): '',
            Metadata.MESH_TECH_PROVIDER_CODE.key(): 'LX',
            Metadata.MESH_TRACE_ID.key(): Mesh.context().get_trace_id(),
            Metadata.MESH_TOKEN.key(): 'x',
            Metadata.MESH_SESSION_ID.key(): 'session_id_001',
            Metadata.MESH_TARGET_NODE_ID.key(): 'LX0000000000000',
        })
        for index in range(1):
            wb = b"ABC" * 30000
            session.push(wb, {}, "1")
            rb = session2.pop(1000, "1")
            if wb.decode('utf-8') != rb.decode('utf-8'):
                log.info("PUSH-POP读写不匹配")
            else:
                log.info("PUSH-POP读写匹配")

            session.push(wb, {}, "1")
            rb = session2.peek("1")
            if wb.decode('utf-8') != rb.decode('utf-8'):
                log.info("PUSH-PEEK读写不匹配")
            else:
                log.info("PUSH-PEEK读写匹配")

        session.release(0)
        session2.release(0)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import random
from typing import Any, List, Dict, Callable

import grpc
import mesh.tool as tool
from mesh.grpx.interceptor import GrpcInterceptor
from mesh.grpx.marshaller import GrpcMarshaller


class GrpcChannels:

    def __init__(self):
        self.channels: Dict[str, GrpcMultiplexChannel] = {}
        self.marshaller = GrpcMarshaller()
        self.serializer = self.marshaller.serialize
        self.deserializer = self.marshaller.deserialize
        self.interceptor = GrpcInterceptor()
        self.path = "/mesh-rpc/v1"
        self.streams: Dict[str, GrpcMultiplexStream] = dict()

    def __exit__(self, exc_type, exc_val, exc_tb):
        for _, channel in self.channels.items():
            channel.close()

    def create_if_absent(self, address: str) -> "GrpcMultiplexChannel":
        channel = self.make_channel(address)
        if channel.counter > 100:
            self.channels.__delitem__(address)
            channel.close()
            channel = self.make_channel(address)
        return channel

    def make_channel(self, address: str) -> "GrpcMultiplexChannel":
        channel = self.channels.get(address, None)
        if not channel:
            self.channels[address] = channel = GrpcMultiplexChannel(address, self.interceptor)
        return channel

    def create_stream(self, address: str) -> grpc.StreamStreamMultiCallable:
        channel = self.create_if_absent(address)
        return channel.stream_stream(self.path, self.serializer, self.deserializer)

    def unary(self, address: str, inbound: bytes, timeout: Any, metadata: Any) -> bytes:
        channel = self.create_if_absent(address)
        conn: grpc.UnaryUnaryMultiCallable = channel.unary_unary(self.path, self.serializer, self.deserializer)
        return conn(inbound, timeout, metadata, None, True, False)

    def stream(self, address: str, inbound: bytes, timeout: Any, metadata: Any) -> bytes:
        stream = self.streams.get(address, None)
        if stream is None:
            stream = self.streams[address] = GrpcMultiplexStream(address, lambda addr: self.create_stream(addr))
        return stream.next(inbound, timeout, metadata)


class GrpcMultiplexStream:

    def __init__(self, address: str, streamer: Callable[[str], grpc.StreamStreamMultiCallable]):
        self.address = address
        self.stream = streamer(address)
        self.streamer = streamer

    def next(self, inbound: bytes, timeout: Any, metadata: Any) -> bytes:
        return self.stream(inbound, timeout, metadata, None, True, False)


class GrpcMultiplexChannel(grpc.Channel):
    """
    https://github.com/grpc/grpc/blob/598e171746576c5398388a4c170ddf3c8d72b80a/include/grpc/impl/codegen/grpc_types.h#L170
    """

    def close(self):
        for _, channel in self.channels:
            channel.close()

    def subscribe(self, callback, try_to_connect=False):
        return self.select().subscribe(callback, try_to_connect)

    def unsubscribe(self, callback):
        return self.select().unsubscribe(callback)

    def unary_unary(self, method, request_serializer=None, response_deserializer=None):
        return self.select().unary_unary(method, request_serializer, response_deserializer)

    def unary_stream(self, method, request_serializer=None, response_deserializer=None):
        return self.select().unary_stream(method, request_serializer, response_deserializer)

    def stream_unary(self, method, request_serializer=None, response_deserializer=None):
        return self.select().stream_unary(method, request_serializer, response_deserializer)

    def stream_stream(self, method, request_serializer=None, response_deserializer=None):
        return self.select().stream_stream(method, request_serializer, response_deserializer)

    def __enter__(self):
        for _, channel in self.channels:
            channel.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        for _, channel in self.channels:
            channel.__exit__(exc_type, exc_val, exc_tb)

    def __init__(self, address: str, *interceptors):
        self.counter = 0
        self.address = address
        self.interceptors = interceptors
        self.channels: List[grpc.Channel] = []
        self.compression_options = {
            "none": grpc.Compression.NoCompression,
            "gzip": grpc.Compression.Gzip,
        }

    def default_compression(self) -> Any:
        return self.compression_options["gzip"]

    def reset(self, channel: grpc.Channel):
        self.channels.append(channel)

    def select(self) -> grpc.Channel:
        """
        grpc.ssl_target_name_override
        grpc.default_authority
        https://github.com/grpc/grpc/blob/master/include/grpc/impl/codegen/compression_types.h
        https://github.com/grpc/grpc/blob/master/include/grpc/impl/codegen/grpc_types.h
        """
        if self.channels.__len__() > 0:
            return self.channels[random.randint(0, self.channels.__len__() - 1)]

        for _ in range(tool.get_max_channels()):
            options = self.default_options()
            channel = grpc.insecure_channel(self.address, options, compression=self.default_compression())
            self.channels.append(grpc.intercept_channel(channel, *self.interceptors))

        return self.channels[random.randint(0, self.channels.__len__() - 1)]

    def default_options(self):
        return [
            ("grpc.enable_retries", True),
            ("grpc.default_compression_algorithm", self.default_compression()),
            ("grpc.client_idle_timeout_ms", 1000 * 60 * 3),
            ("grpc.max_send_message_length", 1 << 30),
            ("grpc.max_receive_message_length", 1 << 30),
            ("grpc.max_connection_age_ms", 1000 * 60 * 12),
            ("grpc.max_connection_age_grace_ms", 1000 * 60 * 4),
            ("grpc.per_message_compression", True),
            ("grpc.per_message_decompression", True),
            ("grpc.enable_deadline_checking", True),
            ("grpc.keepalive_time_ms", 1000 * 60 * 2),
            ("grpc.keepalive_timeout_ms", 1000 * 12),
        ]

    def incr(self):
        self.counter += 1



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import unittest
from abc import ABC, abstractmethod
from typing import Any

from mesh.kinds import Principal
from mesh.macro import mpi, mps
from mesh.mpc import Mesh


class Service(ABC):

    @abstractmethod
    @mpi(name='Service.foo')
    def foo(self, hi: str) -> str:
        pass


@mps
class Implement(Service):

    def foo(self, hi: str) -> str:
        return f'I am {str}'


class TestServiceInvoke(unittest.TestCase):

    @mpi
    @property
    def service(self) -> Service:
        return self.service

    def test_foo(self):
        ret = self.service.foo("Terminator")
        self.assertEqual(ret, "I am Terminator")

    def test_context_safe(self):
        Mesh.context_safe(
            lambda: TestServiceInvoke.phase(
                lambda: TestServiceInvoke.phase(lambda: TestServiceInvoke.phase(lambda: print(-1)))))
        print(len(Mesh.context().get_principals()))

    @staticmethod
    def phase(fn: Any):
        Mesh.context().get_principals().append(Principal())
        Mesh.context_safe(fn)
        print(len(Mesh.context().get_principals()))


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.log.nop import NopLogger
from mesh.log.types import Level, Logger

__all__ = ("Level", "Logger")

mog = NopLogger()


def init():
    """ init function """
    pass


def info(fmt: str, *args: object):
    mog.info(fmt, args)


def warn(fmt: str, *args: object):
    mog.warn(fmt, args)


def error(fmt: str, *args: object):
    mog.error(fmt, args)


def debug(fmt: str, *args: object):
    mog.debug(fmt, args)


def fatal(fmt: str, *args: object):
    mog.fatal(fmt, args)


def stack(fmt: str, *args: object):
    mog.debug(fmt, args)


def level(lev: Level):
    mog.level(lev)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import io
from abc import ABC, abstractmethod
from enum import Enum


def atoi(level: "Level") -> int:
    if isinstance(level, int):
        return level
    return int(level.code)


class Level(Enum):
    FATAL = 1
    ERROR = 2 | atoi(FATAL)
    WARN = 4 | atoi(ERROR)
    INFO = 8 | atoi(WARN)
    DEBUG = 16 | atoi(INFO)
    STACK = 32 | atoi(DEBUG)
    ALL = atoi(FATAL) | atoi(ERROR) | atoi(WARN) | atoi(INFO) | atoi(DEBUG) | atoi(STACK)

    def __init__(self, code: int) -> None:
        self.code = code

    def __str__(self) -> str:
        if Level.FATAL.match(self):
            return 'fatal'
        if Level.ERROR.match(self):
            return 'error'
        if Level.WARN.match(self):
            return 'warn'
        if Level.INFO.match(self):
            return 'info'
        if Level.DEBUG.match(self):
            return 'debug'
        if Level.STACK.match(self):
            return 'stack'
        return 'all'

    def match(self, level: "Level") -> bool:
        return self.code == (self.code & level.code)

    def match_name(self, level: str) -> bool:
        return self.name == level


class Logger(ABC):

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def info(self, fmt: str, *args: object):
        pass

    @abstractmethod
    def warn(self, fmt: str, *args: object):
        pass

    @abstractmethod
    def error(self, fmt: str, *args: object):
        pass

    @abstractmethod
    def debug(self, fmt: str, *args: object):
        pass

    @abstractmethod
    def fatal(self, fmt: str, *args: object):
        pass

    @abstractmethod
    def stack(self, fmt: str, *args: object):
        pass

    @abstractmethod
    def writer(self) -> io.BytesIO:
        pass

    @abstractmethod
    def level(self, level: Level):
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import Locker


@spi("mesh")
class MeshLocker(Locker):

    def __init__(self):
        self.locker = ServiceProxy.default_proxy(Locker)

    def lock(self, rid: str, timeout: int) -> bool:
        return self.locker.lock(rid, timeout)

    def unlock(self, rid: str):
        return self.locker.unlock(rid)

    def read_lock(self, rid: str, timeout: int) -> bool:
        return self.locker.read_lock(rid, timeout)

    def read_unlock(self, rid: str):
        return self.locker.read_unlock(rid)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Dict

from mesh.macro import index, serializable


@serializable
class Document:

    def __init__(self, metadata: Dict[str, str], content: str, timestamp: int):
        self.metadata = metadata
        self.content = content
        self.timestamp = timestamp

    @index(0)
    def metadata(self) -> Dict[str, str]:
        return self.metadata

    @index(5)
    def content(self) -> str:
        return self.content

    @index(10)
    def timestamp(self) -> int:
        return self.timestamp



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any

from mesh.cause import TimeoutException
from mesh.kinds import Reference
from mesh.macro import spi
from mesh.mpc.filter import Filter, Invoker, Invocation, CONSUMER
from mesh.mpc.invoker import Execution


@spi(name="robust", pattern=CONSUMER, priority=0x7fffffff)
class RobustFilter(Filter):

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        execution: Execution[Reference] = invocation.get_execution()
        retries: int = min(execution.schema().retries, 3)
        for _ in range(retries - 1):
            try:
                return invoker.run(invocation)
            except Exception as e:
                if not self.can_retry(e):
                    raise e
                if self.should_delay(e):
                    continue
        return invoker.run(invocation)

    @staticmethod
    def can_retry(e: Exception) -> bool:
        """ Can retry only when timeout occurred """
        return isinstance(e, TimeoutException)

    @staticmethod
    def should_delay(e: Exception) -> bool:
        """ Should delay """
        return False



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any, Dict, Type

from mesh.macro import spi, T
from mesh.prsim import Dispatcher


@spi("mesh")
class MeshDispatcher(Dispatcher):

    def reference(self, mpi: Type[T]) -> T:
        pass

    def invoke(self, urn: str, param: Dict[str, Any]) -> Any:
        pass

    def invoke_generic(self, urn: str, param: Any) -> Any:
        pass



import time

import mesh.log as log
from mesh.cause import CompatibleException


class UUID(object):
    def __init__(self, worker_id, data_center_id):
        self.worker_id = worker_id
        self.data_center_id = data_center_id
        # stats
        self.ids_generated = 0

        # Tue, 21 Mar 2006 20:50:14.000 GMT
        self.epoch = 1142974214000
        self.sequence = 0
        self.worker_id_bits = 5
        self.data_center_id_bits = 5
        self.max_worker_id = -1 ^ (-1 << self.worker_id_bits)
        self.max_data_center_id = -1 ^ (-1 << self.data_center_id_bits)
        self.sequence_bits = 12

        self.worker_id_shift = self.sequence_bits
        self.data_center_id_shift = self.sequence_bits + self.worker_id_bits
        self.timestamp_left_shift = self.sequence_bits + self.worker_id_bits + self.data_center_id_bits
        self.sequence_mask = -1 ^ (-1 << self.sequence_bits)
        self.last_timestamp = -1

        # Sanity check for worker_id
        if self.worker_id > self.max_worker_id or self.worker_id < 0:
            raise CompatibleException(f"Worker id can't be greater than {self.max_worker_id} or less than 0")

        if self.data_center_id > self.max_data_center_id or self.data_center_id < 0:
            raise CompatibleException(f"Data center id can't be greater than {self.max_data_center_id} or less than 0")

        log.debug(
            f"Timestamp left shift {self.timestamp_left_shift}, data center id bits {self.data_center_id_bits}, "
            f"worker id bits {self.worker_id_bits}, sequence bits {self.sequence_bits}, worker id {self.worker_id}. ")

    @staticmethod
    def _time_gen():
        return int(time.time() * 1000)

    def _till_next_millis(self, last_timestamp):
        timestamp = self._time_gen()
        while timestamp <= last_timestamp:
            timestamp = self._time_gen()

        return timestamp

    def _next_id(self):
        timestamp = self._time_gen()

        if self.last_timestamp > timestamp:
            log.warn(f"Clock is moving backwards. Rejecting request until {self.last_timestamp}")

        if self.last_timestamp == timestamp:
            self.sequence = (self.sequence + 1) & self.sequence_mask
            if self.sequence == 0:
                timestamp = self._till_next_millis(self.last_timestamp)
        else:
            self.sequence = 0

        self.last_timestamp = timestamp

        new_id = ((timestamp - self.epoch) << self.timestamp_left_shift) | (
                self.data_center_id << self.data_center_id_shift) | (
                         self.worker_id << self.worker_id_shift) | self.sequence
        self.ids_generated += 1
        return new_id

    def get_worker_id(self):
        return self.worker_id

    def get_timestamp(self):
        return self._time_gen()

    def new_id(self):
        return self._next_id()

    def get_datacenter_id(self):
        return self.data_center_id



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any, Dict

import mesh.log as log
from mesh.cause import MeshCode, Codeable, MeshException
from mesh.context import Mesh
from mesh.macro import spi
from mesh.mpc.digest import Digest
from mesh.mpc.filter import Filter, Invoker, Invocation, PROVIDER


@spi(name="provider", pattern=PROVIDER, priority=0x7fffffff)
class ProviderFilter(Filter):

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not attachments:
            attachments = {}
        Mesh.context().decode(attachments)

        digest = Digest()
        try:
            ret = invoker.run(invocation)
            digest.write("P", MeshCode.SUCCESS.get_code())
            return ret
        except BaseException as e:
            if isinstance(e, Codeable):
                digest.write("P", e.get_code())
            else:
                digest.write("P", MeshCode.SYSTEM_ERROR.get_code())
            if isinstance(e, MeshException):
                log.error(f"{digest.trace_id},{Mesh.context().get_urn()},{e.get_message}")
            raise e



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import inspect
import io
import traceback
from datetime import datetime
from io import StringIO

from mesh.environ import Mode
from mesh.log.types import Logger, Level


def stack_tuple() -> (str, str):
    if Mode.Nolog:
        return "", ""
    frame = inspect.stack()[4]
    index = frame.filename.rindex("/")
    name = frame.filename[0 if index < 0 else index + 1:]
    return name, frame.lineno


class NopLogger(Logger):

    def __init__(self):
        self.__level__ = Level.INFO

    def name(self) -> str:
        return 'mesh'

    def info(self, fmt: str, *args: object):
        self.write(Level.INFO, fmt, *args)

    def warn(self, fmt: str, *args: object):
        self.write(Level.WARN, fmt, *args)

    def error(self, fmt: str, *args: object):
        self.write(Level.ERROR, fmt, *args)

    def debug(self, fmt: str, *args: object):
        self.write(Level.DEBUG, fmt, *args)

    def fatal(self, fmt: str, *args: object):
        self.write(Level.FATAL, fmt, *args)

    def stack(self, fmt: str, *args: object):
        self.write(Level.STACK, fmt, *args)

    def writer(self) -> io.BytesIO:
        return io.BytesIO()

    def level(self, level: Level):
        self.__level__ = level

    def write(self, level: Level, fmt: str, args):
        if not level.match(self.__level__):
            return

        # discard log
        if Mode.Nolog.enable():
            return

        if not args or len(args) < 1:
            self.do_write(level, fmt)
            return

        if len(args) == 1 and isinstance(args[0], BaseException):
            self.do_write(level, fmt, args[0])
            return

        frags = (fmt if fmt else "").split("{}")
        placeholders = []
        throwable = False
        for arg in args:
            if isinstance(arg, BaseException):
                throwable = True
                continue
            if isinstance(arg, str):
                placeholders.append(arg)
                continue
            placeholders.append(f"{arg}")

        msg = StringIO()
        for idx in range(len(frags)):
            msg.write(frags[idx])
            if len(placeholders) > idx:
                msg.write(placeholders[idx])
                continue
            if idx < len(frags) - 1:
                msg.write('{}')

        self.do_write(level, msg.getvalue(), throwable)

    def do_write(self, level: Level, msg: str, e: bool = False):
        print(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} {level.name} {msg}{self.stack_tuple()}")
        if e:
            print(traceback.format_exc())

    @staticmethod
    def stack_tuple() -> str:
        if not Mode.Metrics.enable():
            return ""
        frame = inspect.stack()[4]
        index = frame.filename.rindex("/")
        name = frame.filename[0 if index < 0 else index + 1:]
        return f" {name}:{frame.lineno}"



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


import os
from glob import glob
from itertools import chain
from os.path import join

import pytest

import mesh.log as log


def call_pytest_subprocess(*tests, **kwargs):
    global _ctr

    _ctr += 1
    file_name = 'test_result.%d.xml' % _ctr
    if os.path.isfile(file_name):
        os.unlink(file_name)
    env = {}
    args = [
        '--verbose',
        '--cov-append',
        '--cov-report=',
        '--cov', 'spyne',
        '--tb=line',
        '--junitxml=%s' % file_name
    ]
    args.extend('--{0}={1}'.format(k, v) for k, v in kwargs.items())
    return call_test(pytest.main, args, tests, env)


def call_test(self, fn, a, tests, env={}):
    from multiprocessing import Process, Queue

    tests_dir = os.path.dirname(self.__file__)
    if len(tests) > 0:
        a.extend(chain(*[glob(join(tests_dir, test)) for test in tests]))

    queue = Queue()
    p = Process(target=_wrapper(fn), args=[a, queue, env])
    p.start()
    p.join()

    ret = queue.get()
    if ret == 0:
        log.info(tests or a, "OK")
    else:
        log.info(tests or a, "FAIL")

    return ret


def _wrapper(fn):
    import traceback

    def _(args, queue, env):
        log.info("env:", env)
        for k, v in env.items():
            os.environ[k] = v
        try:
            ret = fn(args)
        except SystemExit as e:
            ret = e.code
        except BaseException as e:
            log.info(traceback.format_exc())
            ret = 1

        queue.put(ret)

    return _


class RunTests:
    def run_tests(self):
        ret = 0
        tests = [
            'interface', 'model', 'multipython', 'protocol', 'util',
            'interop/test_pyramid.py',
            'interop/test_soap_client_http_twisted.py',
            'test_null_server.py',
            'test_service.py',
            'test_soft_validation.py',
            'test_sqlalchemy.py',
            'test_sqlalchemy_deprecated.py',
        ]

        log.info("Test stage 1: Unit tests")
        ret = call_pytest_subprocess(*tests, capture=self.capture) or ret

        log.info("\nTest stage 2: End-to-end tests")
        ret = call_pytest_subprocess('interop/test_httprpc.py', capture=self.capture) or ret
        ret = call_pytest_subprocess('interop/test_soap_client_http.py', capture=self.capture) or ret
        ret = call_pytest_subprocess('interop/test_soap_client_zeromq.py', capture=self.capture) or ret
        raise SystemExit(ret)


__all__ = (
    RunTests
)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

# import psutil
# import socket
# from prometheus_client.metrics_core import CounterMetricFamily, GaugeMetricFamily
#
# from prometheus_client.registry import Collector
#
# from typing import Iterable
#
# from prometheus_client import Metric
#
#
# class CPUCollector(Collector):
#
#     def __init__(self):
#         self.hostname = socket.gethostname()
#
#     def collect(self) -> Iterable[Metric]:
#         total_cpu_times = psutil.cpu_times(True)
#         cpu_counter = CounterMetricFamily("node_cpu_seconds_total", "Seconds the CPUs spent in each mode.",
#                                           labels=["cpu", "mode", "host"])
#         for idx, per_cpu in enumerate(total_cpu_times):
#             print(per_cpu)
#             for k, v in per_cpu._asdict().items():
#                 cpu_counter.add_metric([idx, k, self.hostname], value=v)
#         return [cpu_counter]
#
#
# class DiskStatsCollector(Collector):
#
#     def __init__(self):
#         self.hostname = socket.gethostname()
#
#     def collect(self) -> Iterable[Metric]:
#         read_completed_counter = CounterMetricFamily("node_memory_reads_completed_total",
#                                                      "The total number of reads completed successfully.",
#                                                      labels=["device", "host"])
#         read_bytes_total = CounterMetricFamily("node_memory_read_bytes_total",
#                                                "The total number of bytes read successfully.",
#                                                labels=["device", "host"])
#         writes_completed_total = CounterMetricFamily("node_memory_writes_completed_total",
#                                                      "The total number of writes completed successfully.",
#                                                      labels=["device", "host"])
#         written_bytes_total = CounterMetricFamily("node_memory_written_bytes_total",
#                                                   "The total number of bytes written successfully.",
#                                                   labels=["device", "host"])
#         # io_now = Gauge("node_memory_io_now",
#         #               "The number of I/Os currently in progress.",
#         #               ["device", "host"])
#         # io_time_seconds_total = Counter("node_memory_io_time_seconds_total",
#         #                                "Total seconds spent doing I/Os.",
#         #                                ["device", "host"])
#         di_counters = psutil.disk_io_counters(perdisk=True)
#         if not di_counters:
#             return []
#         for k, v in di_counters.items():
#             read_completed_counter.add_metric([k, self.hostname], value=v.read_count)
#             read_bytes_total.add_metric([k, self.hostname], value=v.read_bytes)
#             writes_completed_total.add_metric([k, self.hostname], value=v.write_count)
#             written_bytes_total.add_metric([k, self.hostname], value=v.write_bytes)
#             # io_now.labels(k, self.hostname).inc(v["read_count"])
#             # io_time_seconds_total.labels(k, self.hostname).inc(v["busy_time"])
#         return [read_completed_counter, read_bytes_total, writes_completed_total, written_bytes_total]
#
#
# class MemInfoCollector(Collector):
#     def __init__(self):
#         self.hostname = socket.gethostname()
#
#     def collect(self) -> Iterable[Metric]:
#         available = GaugeMetricFamily("node_memory_available_bytes",
#                                       "Memory information field available.",
#                                       labels=["host"])
#         total = CounterMetricFamily("node_memory_total_bytes",
#                                     "Memory information field total.",
#                                     labels=["host"])
#         active = GaugeMetricFamily("node_memory_active_bytes",
#                                    "Memory information field active.",
#                                    labels=["host"])
#         memory_view = psutil.virtual_memory()
#         if not memoryview:
#             return []
#         available.add_metric([self.hostname], value=memory_view.available)
#         total.add_metric([self.hostname], value=memory_view.total)
#         active.add_metric([self.hostname], value=memory_view.active)
#         ret = []
#         ret.extend([available, active, total])
#         return ret
#
#
# class NetworkCollector(Collector):
#     def __init__(self):
#         self.hostname = socket.gethostname()
#
#     def collect(self) -> Iterable[Metric]:
#         ni_counters = psutil.net_io_counters(pernic=True)
#         if not ni_counters:
#             return []
#         recv = CounterMetricFamily("node_network_receive",
#                                    "Network device statistic receive.",
#                                    labels=["device", "host"])
#         transmit = GaugeMetricFamily("node_network_transmit",
#                                      "Network device statistic transmit.",
#                                      labels=["device", "host"])
#         for k, v in ni_counters.items():
#             recv.add_metric([k, self.hostname], value=v.bytes_recv)
#             transmit.add_metric([k, self.hostname], value=v.bytes_sent)
#         ret = []
#         ret.extend([recv, transmit])
#         return ret
#
#
# CPU_COLLECTOR = CPUCollector()
# MEM_COLLECTOR = MemInfoCollector()
# DISK_COLLECTOR = DiskStatsCollector()
# NETWORK_COLLECTOR = NetworkCollector()
#
# if __name__ == "__main__":
#     cpu_collect = CPUCollector()
#     mem_collect = MemInfoCollector()
#     disk_collect = DiskStatsCollector()
#     network_collect = NetworkCollector()
#     for e in cpu_collect.collect():
#
#     for e in mem_collect.collect():
#
#     for e in disk_collect.collect():
#
#     for e in network_collect.collect():
#



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.cause.status import Codeable, MeshCode


class MeshException(RuntimeError, Codeable):

    def __init__(self, code: str, message: str):
        self.code = code
        self.message = message

    def __str__(self):
        return f'{self.code}, {self.message}'

    def get_code(self):
        return self.code

    def get_message(self):
        return self.message


class CompatibleException(MeshException):

    def __init__(self, message: str):
        super().__init__(MeshCode.COMPATIBLE_ERROR.get_code(), message)


class NotFoundException(MeshException):

    def __init__(self, message: str):
        super().__init__(MeshCode.NOT_FOUND.get_code(), message)


class ValidationException(MeshException):
    """
    Validate error
    """

    def __init__(self, message: str):
        super().__init__(MeshCode.VALIDATE_ERROR.get_code(), message)


class TimeoutException(MeshException):
    """
    TimeoutException
    """

    def __init__(self, message: str):
        super().__init__(MeshCode.TIMEOUT_ERROR.get_code(), message)


class NoProviderException(MeshException):
    """
    NoProviderException
    """

    def __init__(self, message: str):
        super().__init__(MeshCode.NO_PROVIDER_ERROR.get_code(), message)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import functools
import inspect
from typing import Any, TypeVar, Dict, List, Type

import mesh.log as log

T = TypeVar('T')
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class Species:
    macro = Any
    name = ''
    kind = Type[Any]
    metadata = Any
    trait = Type[Any]
    abstract = False


class Ark:
    def __init__(self):
        self.tinder: Dict[str, List[Species]] = dict()

    def register(self, macro: Any, name: str, kind: type, metadata: Any):
        """
        Register a annotated object as a metadata.
        :param macro: annotated decorator
        :param name: object name
        :param kind: object type
        :param metadata: decorated object
        :return: none
        """
        traits = [kind]
        if hasattr(kind, '__bases__'):
            for base in kind.__bases__:
                traits.append(base)

        for trait in traits:
            key = f'{macro.__name__}-{name}-{trait.__name__}'
            specs = self.tinder.get(key, [])
            for spec in specs:
                if spec.kind == kind:
                    log.warn(f'Object of {trait.__name__} named {name} has been register already.')

            spec = Species()
            spec.macro = macro
            spec.name = name
            spec.kind = kind
            spec.metadata = metadata
            spec.trait = trait
            spec.abstract = False
            if inspect.isabstract(kind):
                spec.abstract = True
            specs.append(spec)

            self.tinder[key] = specs

    def unregister(self, macro: Any, trait: type):
        matches = []
        for key, species in self.tinder.items():
            if species.macro == macro and species.trait == trait:
                matches.append(key)
        for _, key in matches:
            del self.tinder[key]

    def export(self, macro: Any) -> Dict[Type[T], Dict[str, List[Species]]]:
        matches: Dict[type[T], Dict[str, List[Species]]] = dict()
        for _, specs in self.tinder.items():
            for spec in specs:
                if spec.macro != macro or spec.abstract:
                    continue
                if not matches.get(spec.trait):
                    matches[spec.trait] = dict()
                if not matches.get(spec.trait).get(spec.name):
                    matches[spec.trait][spec.name] = []
                matches[spec.trait][spec.name].append(spec)

        return matches

    def providers(self, macro: Any, kind: Type[T]) -> List[Species]:
        matches = list()
        for _, specs in self.export(macro).get(kind, {}).items():
            for spec in specs:
                matches.append(spec)

        return matches

    def trait(self, macro: Any, kind: Type[T], name: str) -> List[Type[T]]:
        key = f'{macro.__name__}-{name}-{kind.__name__}'
        pvs: List[Type[T]] = []
        for spec in self.tinder.get(key, []):
            if not spec.abstract:
                pvs.append(spec.kind)

        return pvs

    @staticmethod
    def get_declared_class(method: Any):
        if isinstance(method, functools.partial):
            return Ark.get_declared_class(method.func)
        if inspect.ismethod(method) or (
                inspect.isbuiltin(method) and hasattr(method, '__self__') and hasattr(method.__self__, '__class__')):
            for cls in inspect.getmro(method.__self__.__class__):
                if method.__name__ in cls.__dict__:
                    return cls
            method = getattr(method, '__func__', method)  # fallback to __qualname__ parsing
        if inspect.isfunction(method):
            class_name = method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0]
            cls = getattr(inspect.getmodule(method), class_name, None)
            if not cls:
                cls = method.__globals__.get(class_name, None)
            if isinstance(cls, type):
                return cls
        return getattr(method, '__objclass__', None)  # handle special descriptor objects


ark = Ark()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod
from typing import Generic

from mesh.macro import mpi, spi, T, A


@spi("mesh")
class Endpoint(ABC):
    """
    Like subscriber but synchronized and has return value. For any extension.
    """

    @abstractmethod
    @mpi("${mesh.uname}")
    def fuzzy(self, buff: bytes) -> bytes:
        """
        Invoke endpoint.
        :param buff:
        :return:
        """
        pass


@spi("mesh")
class EndpointSticker(ABC, Generic[T, A]):
    """
    Like subscriber but synchronized and has return value. For any extension.
    """

    @abstractmethod
    def stick(self, varg: T) -> A:
        """
        Invoke endpoint.
        :param varg:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import abstractmethod
from typing import Any


class Context:

    @abstractmethod
    def inject_properties(self, properties: bytes):
        pass

    @abstractmethod
    def register_processor(self, processor: Any):
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import abstractmethod, ABC
from typing import Generic, Dict

from mesh.macro import spi, mpi, T


@spi("mesh")
class Transport(ABC, Generic[T]):
    """
    Private compute data channel in async and blocking mode.
    """

    MESH = "mesh"
    GRPC = "grpc"

    @abstractmethod
    @mpi("mesh.chan.open")
    def open(self, session_id: str, metadata: Dict[str, str]) -> "Session":
        """
        Open a channel session.
        :param session_id:  node id or inst id
        :param metadata channel metadata
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.chan.close")
    def close(self, timeout: int):
        """
        Close the channel.
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.chan.roundtrip")
    def roundtrip(self, payload: bytes, metadata: Dict[str, str]) -> bytes:
        """
        Roundtrip with the channel.
        :param payload:
        :param metadata:
        :return:
        """
        pass


@spi("mesh")
class Session(ABC, Generic[T]):
    """
    Remote queue in async and blocking mode.
    """

    @abstractmethod
    @mpi("mesh.chan.peek")
    def peek(self, topic: str = "") -> bytes:
        """
        Retrieves, but does not remove, the head of this queue,
        or returns None if this queue is empty.
        :param topic: message topic
        :return: the head of this queue, or None if this queue is empty
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="mesh.chan.pop", timeout=120 * 1000 * 30)
    def pop(self, timeout: int, topic: str = "") -> bytes:
        """
        Retrieves and removes the head of this queue,
        or returns None if this queue is empty.
        :param timeout: timeout in mills.
        :param topic: message topic
        :return: the head of this queue, or None if this queue is empty
        """
        pass

    @abstractmethod
    @mpi("mesh.chan.push", timeout=120 * 1000 * 30)
    def push(self, payload: bytes, metadata: Dict[str, str], topic: str = ""):
        """
        Inserts the specified element into this queue if it is possible to do
        so immediately without violating capacity restrictions.
        When using a capacity-restricted queue, this method is generally
        preferable to add, which can fail to insert an element only
        by throwing an exception.
        :param payload: message payload
        :param metadata: Message metadata
        :param topic: message topic
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.chan.release")
    def release(self, timeout: int, topic: str = ""):
        """
        Close the channel session.
        :param timeout:
        :param topic: message topic
        :return:
        """
        pass

    async def pick(self, topic: str = "") -> bytes:
        """ Peek the instance """
        return self.peek(topic)

    async def poll(self, timeout: int, topic: str = "") -> bytes:
        """ Pop the instance """
        return self.pop(timeout, topic)

    async def offer(self, payload: bytes, metadata: Dict[str, str], topic: str = ""):
        """ Push the instance """
        self.push(payload, metadata, topic)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.system.mesh_builtin import MeshBuiltin
from mesh.system.mesh_cache import MeshCache
from mesh.system.mesh_cluster import MeshCluster
from mesh.system.mesh_datahouse import MeshDataHouse
from mesh.system.mesh_dispatcher import MeshDispatcher
from mesh.system.mesh_endpoint import MeshEndpoint
from mesh.system.mesh_evaluator import MeshEvaluator
from mesh.system.mesh_hodor import MeshHodor
from mesh.system.mesh_locker import MeshLocker
from mesh.system.mesh_network import MeshNetwork
from mesh.system.mesh_publisher import MeshPublisher
from mesh.system.mesh_registry import MeshRegistry
from mesh.system.mesh_runtime_hook import MeshRuntimeHook
from mesh.system.mesh_scheduler import MeshScheduler
from mesh.system.mesh_transport import MeshTransport, MeshSession

__all__ = (
    "MeshBuiltin",
    "MeshCache",
    "MeshCluster",
    "MeshDataHouse",
    "MeshDispatcher",
    "MeshEndpoint",
    "MeshEvaluator",
    "MeshHodor",
    "MeshLocker",
    "MeshNetwork",
    "MeshPublisher",
    "MeshRegistry",
    "MeshRuntimeHook",
    "MeshScheduler",
    "MeshTransport",
    "MeshSession",
)


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import mesh.tool as tool
from mesh.macro import ServiceLoader
from mesh.mpc import Provider
from mesh.prsim import RuntimeHook


class Mooter(RuntimeHook):
    DEFAULT_PORT = 8864

    def __init__(self):
        self.http1: Provider = None
        self.http2: Provider = None

    def start(self):
        hooks = ServiceLoader.load(RuntimeHook).list('')
        for hook in hooks:
            hook.start()

        self.http1: Provider = ServiceLoader.load(Provider).get("http")
        self.http2: Provider = ServiceLoader.load(Provider).get("grpc")
        self.http1.start(f"0.0.0.0:{tool.get_mesh_runtime().port + 1}", None)
        self.http2.start(f"0.0.0.0:{tool.get_mesh_runtime().port}", None)

    def stop(self):
        if self.http1 is not None:
            self.http1.close()
        if self.http2 is not None:
            self.http2.close()
        hooks = ServiceLoader.load(RuntimeHook).list('')
        for hook in hooks:
            hook.stop()

    def refresh(self):
        self.stop()
        self.start()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.grpx.bindservice import GrpcBindableService
from mesh.grpx.channels import GrpcChannels
from mesh.grpx.consumer import GrpcConsumer
from mesh.grpx.interceptor import GrpcInterceptor
from mesh.grpx.marshaller import GrpcMarshaller
from mesh.grpx.provider import GrpcProvider

MPC = "/mesh-rpc/v1"

__all__ = (
    "GrpcBindableService",
    "GrpcChannels",
    "GrpcConsumer",
    "GrpcInterceptor",
    "GrpcMarshaller",
    "GrpcProvider"
)


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from concurrent.futures import Future
from typing import Any

import grpc
import mesh.codec.tools as tools
import mesh.tool as tool
import requests
from mesh.cause import MeshException, MeshCode
from mesh.context import Mesh
from mesh.grpx.channels import GrpcChannels
from mesh.kinds import Reference
from mesh.mpc.invoker import Execution
from mesh.ptp.y_pb2 import PeekInbound, PopInbound, PushInbound, ReleaseInbound, TransportOutbound


class BfiaBinding:
    def __init__(self, urn: str, args: Any):
        if urn.startswith("peek.chan.mesh"):
            self.path = "/org.ppc.ptp.PrivateTransferTransport/peek"
            self.h1path = "/v1/interconn/chan/peek"
            self.inbound = PeekInbound()
            self.inbound.topic = args.topic
            self.outbound = TransportOutbound
        elif urn.startswith("pop.chan.mesh"):
            self.path = "/org.ppc.ptp.PrivateTransferTransport/pop"
            self.h1path = "/v1/interconn/chan/pop"
            self.inbound = PopInbound()
            self.inbound.topic = args.topic
            self.inbound.timeout = args.timeout
            self.outbound = TransportOutbound
        elif urn.startswith("push.chan.mesh"):
            self.path = "/org.ppc.ptp.PrivateTransferTransport/push"
            self.h1path = "/v1/interconn/chan/push"
            self.inbound = PushInbound()
            self.inbound.topic = args.topic
            self.inbound.payload = args.payload
            if args.metadata:
                for k, v in args.metadata.items():
                    self.inbound.metadata[k] = v
            self.outbound = TransportOutbound
        elif urn.startswith("release.chan.mesh"):
            self.path = "/org.ppc.ptp.PrivateTransferTransport/release"
            self.h1path = "/v1/interconn/chan/release"
            self.inbound = ReleaseInbound()
            self.inbound.topic = args.topic
            self.inbound.timeout = args.timeout
            self.outbound = TransportOutbound
        else:
            self.path = ""
            self.inbound = None
            self.outbound = None

    def encode(self, _: bytes):
        return self.inbound.SerializeToString()

    def decode(self, buffer: bytes):
        out = self.outbound()
        out.ParseFromString(buffer)
        v = "{" + f'"code":"{out.code}","message":"{out.message}","content":"{tools.b64encode(out.payload)}"' + "}"
        return v.encode('utf-8')


def bfia_consume(address: str, binding: BfiaBinding, execution: Execution[Reference], inbound: bytes,
                 channel: GrpcChannels) -> Future:
    if 'grpc' != tool.get_bfia_proto():
        return bfia_consume_h11(address, binding, execution)

    urn = Mesh.context().get_urn()
    try:
        Mesh.context().rewrite_urn("")
        ct = Mesh.context().get_attribute(Mesh.TIMEOUT)
        timeout = (ct if ct else execution.schema().timeout) / 1000
        channel = channel.create_if_absent(address)
        conn: grpc.UnaryUnaryMultiCallable = channel.unary_unary(binding.path, binding.encode, binding.decode)
        outbound = conn(inbound, timeout, [], None, True, False)
        future = Future()
        future.set_result(outbound)
    finally:
        Mesh.context().rewrite_urn(urn)
    return future


session = requests.Session()


def bfia_consume_h11(address: str, binding: BfiaBinding, execution: Execution[Reference]) -> Future:
    urn = Mesh.context().get_urn()
    try:
        Mesh.context().rewrite_urn("")
        ct = Mesh.context().get_attribute(Mesh.TIMEOUT)
        timeout = (ct if ct else execution.schema().timeout) / 1000
        headers = {'content-type': 'application/x-protobuf'}
        for (k, v) in Mesh.context().get_attachments().items():
            if k.startswith('x-ptp-'):
                headers[k] = v
        rs = session.post(url=f"http://{address}{binding.h1path}", data=binding.encode(b''), timeout=timeout,
                          headers=headers)

        if rs.status_code != 200:
            raise MeshException(MeshCode.SYSTEM_ERROR.get_code(), f"{binding.h1path} {rs.status_code} {rs.text}")
        outbound = binding.decode(rs.content)
        future = Future()
        future.set_result(outbound)
    finally:
        Mesh.context().rewrite_urn(urn)
    return future



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: x.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
    name='x.proto',
    package='org.ppc.ptp',
    syntax='proto3',
    serialized_options=b'\n\013io.bfia.ptpZ\031github.com/be-io/mesh/ptp',
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x07x.proto\x12\x0borg.ppc.ptp\"\x81\x01\n\x07Inbound\x12\x34\n\x08metadata\x18\x01 \x03(\x0b\x32\".org.ppc.ptp.Inbound.MetadataEntry\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa2\x01\n\x08Outbound\x12\x35\n\x08metadata\x18\x01 \x03(\x0b\x32#.org.ppc.ptp.Outbound.MetadataEntry\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\x12\x0c\n\x04\x63ode\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32\x8e\x01\n\x17PrivateTransferProtocol\x12<\n\ttransport\x12\x14.org.ppc.ptp.Inbound\x1a\x15.org.ppc.ptp.Outbound(\x01\x30\x01\x12\x35\n\x06invoke\x12\x14.org.ppc.ptp.Inbound\x1a\x15.org.ppc.ptp.OutboundB(\n\x0bio.bfia.ptpZ\x19github.com/be-io/mesh/ptpb\x06proto3'
)




_INBOUND_METADATAENTRY = _descriptor.Descriptor(
    name='MetadataEntry',
    full_name='org.ppc.ptp.Inbound.MetadataEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='key', full_name='org.ppc.ptp.Inbound.MetadataEntry.key', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='value', full_name='org.ppc.ptp.Inbound.MetadataEntry.value', index=1,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=b'8\001',
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=107,
    serialized_end=154,
)

_INBOUND = _descriptor.Descriptor(
    name='Inbound',
    full_name='org.ppc.ptp.Inbound',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='metadata', full_name='org.ppc.ptp.Inbound.metadata', index=0,
            number=1, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='payload', full_name='org.ppc.ptp.Inbound.payload', index=1,
            number=2, type=12, cpp_type=9, label=1,
            has_default_value=False, default_value=b"",
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[_INBOUND_METADATAENTRY, ],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=25,
    serialized_end=154,
)


_OUTBOUND_METADATAENTRY = _descriptor.Descriptor(
    name='MetadataEntry',
    full_name='org.ppc.ptp.Outbound.MetadataEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='key', full_name='org.ppc.ptp.Outbound.MetadataEntry.key', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='value', full_name='org.ppc.ptp.Outbound.MetadataEntry.value', index=1,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=b'8\001',
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=107,
    serialized_end=154,
)

_OUTBOUND = _descriptor.Descriptor(
    name='Outbound',
    full_name='org.ppc.ptp.Outbound',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='metadata', full_name='org.ppc.ptp.Outbound.metadata', index=0,
            number=1, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='payload', full_name='org.ppc.ptp.Outbound.payload', index=1,
            number=2, type=12, cpp_type=9, label=1,
            has_default_value=False, default_value=b"",
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='code', full_name='org.ppc.ptp.Outbound.code', index=2,
            number=3, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='message', full_name='org.ppc.ptp.Outbound.message', index=3,
            number=4, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[_OUTBOUND_METADATAENTRY, ],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=157,
    serialized_end=319,
)

_INBOUND_METADATAENTRY.containing_type = _INBOUND
_INBOUND.fields_by_name['metadata'].message_type = _INBOUND_METADATAENTRY
_OUTBOUND_METADATAENTRY.containing_type = _OUTBOUND
_OUTBOUND.fields_by_name['metadata'].message_type = _OUTBOUND_METADATAENTRY
DESCRIPTOR.message_types_by_name['Inbound'] = _INBOUND
DESCRIPTOR.message_types_by_name['Outbound'] = _OUTBOUND
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Inbound = _reflection.GeneratedProtocolMessageType('Inbound', (_message.Message,), {

    'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
        'DESCRIPTOR' : _INBOUND_METADATAENTRY,
        '__module__' : 'x_pb2'
        # @@protoc_insertion_point(class_scope:org.ppc.ptp.Inbound.MetadataEntry)
    })
    ,
    'DESCRIPTOR' : _INBOUND,
    '__module__' : 'x_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.Inbound)
})
_sym_db.RegisterMessage(Inbound)
_sym_db.RegisterMessage(Inbound.MetadataEntry)

Outbound = _reflection.GeneratedProtocolMessageType('Outbound', (_message.Message,), {

    'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
        'DESCRIPTOR' : _OUTBOUND_METADATAENTRY,
        '__module__' : 'x_pb2'
        # @@protoc_insertion_point(class_scope:org.ppc.ptp.Outbound.MetadataEntry)
    })
    ,
    'DESCRIPTOR' : _OUTBOUND,
    '__module__' : 'x_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.Outbound)
})
_sym_db.RegisterMessage(Outbound)
_sym_db.RegisterMessage(Outbound.MetadataEntry)


DESCRIPTOR._options = None
_INBOUND_METADATAENTRY._options = None
_OUTBOUND_METADATAENTRY._options = None

_PRIVATETRANSFERPROTOCOL = _descriptor.ServiceDescriptor(
    name='PrivateTransferProtocol',
    full_name='org.ppc.ptp.PrivateTransferProtocol',
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=322,
    serialized_end=464,
    methods=[
        _descriptor.MethodDescriptor(
            name='transport',
            full_name='org.ppc.ptp.PrivateTransferProtocol.transport',
            index=0,
            containing_service=None,
            input_type=_INBOUND,
            output_type=_OUTBOUND,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name='invoke',
            full_name='org.ppc.ptp.PrivateTransferProtocol.invoke',
            index=1,
            containing_service=None,
            input_type=_INBOUND,
            output_type=_OUTBOUND,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ])
_sym_db.RegisterServiceDescriptor(_PRIVATETRANSFERPROTOCOL)

DESCRIPTOR.services_by_name['PrivateTransferProtocol'] = _PRIVATETRANSFERPROTOCOL

# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import io
import threading
import time
from typing import List

from mesh import tool, log
from mesh.boost.disruptor import Disruptor, Consumer
from mesh.kinds import Document
from mesh.log.types import Logger, Level
from mesh.macro import mpi, spi
from mesh.mpc import Mesh
from mesh.prsim import DataHouse, Key

DISCARD = Key("mesh.syslog.discard")
BUFFER_SIZE = 10
LIMITER = threading.Semaphore(3)


@spi("remote")
class RemoteLogger(Logger):

    def __init__(self, level: Level):
        self.__level__ = level
        self.disruptor = Disruptor(name='mesh', size=64)
        self.disruptor.register_consumer(DocumentConsumer())

    def name(self) -> str:
        return 'remote'

    def info(self, fmt: str, *args: object):
        self.write(Level.INFO, fmt, *args)

    def warn(self, fmt: str, *args: object):
        self.write(Level.WARN, fmt, *args)

    def error(self, fmt: str, *args: object):
        self.write(Level.ERROR, fmt, *args)

    def debug(self, fmt: str, *args: object):
        self.write(Level.DEBUG, fmt, *args)

    def fatal(self, fmt: str, *args: object):
        self.write(Level.FATAL, fmt, *args)

    def stack(self, fmt: str, *args: object):
        self.write(Level.STACK, fmt, *args)

    def writer(self) -> io.BytesIO:
        return io.BytesIO()

    def level(self, level: Level):
        self.__level__ = level

    def write(self, level: Level, fmt: str, *args):
        if not level.match(self.__level__):
            return
        is_discard = Mesh.context().get_attribute(DISCARD)
        if is_discard is not None and is_discard:
            return
        metadata = {
            "name": "",
            "level": level.name,
            "host": tool.get_hostname(),
            "ip": tool.get_ip(),
            "mesh_runtime": str(tool.get_mesh_runtime()),
            "mesh_trace_id": Mesh.context().get_trace_id(),
            "mesh_span_id": Mesh.context().get_span_id(),
            "mesh_name": tool.get_mesh_name(),
            "mesh_run_mode": str(Mesh.context().get_run_mode())
        }
        document = Document(metadata, fmt, time.time_ns())
        self.disruptor.produce([document])


class DocumentConsumer(Consumer):

    @mpi
    def data_house(self) -> DataHouse:
        """"""
        pass

    def consume(self, elements):
        return Mesh.context_safe(lambda: self.do_consume(elements))

    def do_consume(self, elements):
        Mesh.context().set_attribute(DISCARD, True)
        if len(elements) > BUFFER_SIZE - 1:
            try:
                self.data_house.writes(elements)
            except BaseException as e:
                self.exception(e, elements)
            finally:
                elements.clear()
            return
        if not LIMITER.acquire(True):
            return
        try:
            self.data_house.write(elements)
        except BaseException as e:
            self.exception(e, elements)
        finally:
            LIMITER.release(1)
            elements.clear()

    @staticmethod
    def exception(e: BaseException, event: List[Document]):
        try:
            for document in event:
                ln = document.metadata.get("level", '')
                if Level.ERROR.match_name(ln):
                    log.error(document.content)
                elif Level.WARN.match_name(ln):
                    log.warn(document.content)
                else:
                    log.info(document.content)
            log.error("", e)
        except BaseException as c:
            log.error("", c)
        finally:
            event.clear()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import logging
import random
import time
from threading import Event
from unittest import TestCase

from mesh.boost import Disruptor, BatchConsumer, Consumer
from mesh.boost import RingBuffer, RingBufferLagStats


class TestRingBuffer(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.logger = logging.getLogger(cls.__name__)
        cls.logger.level = logging.DEBUG

    def test_get_set(self):
        ring = RingBuffer(5)
        # assert intial ring state with wrap-around
        for idx in range(10):
            self.assertEqual(None, ring.get(idx))
        # set some elements
        for idx in range(10):
            ring.set(idx, idx)
        # assert values
        for idx in range(5):
            self.assertEqual(idx + 5, ring.get(idx))

    def test_mget_mset(self):
        ring = RingBuffer(10)
        for idx in range(10):
            ring.set(idx, idx)

        ring.mset(0, [9, 9, 9])
        # ring is now [9,9,9,3,4,5,6,7,8,9]
        self.assertEqual([9, 9, 9, 3], ring.mget(0, 4))
        self.assertEqual([9, 9, 3, 4], ring.mget(1, 4))
        self.assertEqual([8, 9, 9, 9, 9, 3], ring.mget(8, 6))
        ring.mset(8, [1, 2, 3, 4])
        # ring is now [3,4,9,3,4,5,6,7,1,2]
        self.assertEqual([3, 4, 9, 3, 4, 5, 6, 7, 1, 2], ring.mget(0, 10))
        self.assertEqual([4, 9, 3, 4, 5, 6, 7, 1, 2, 3], ring.mget(1, 10))
        ring.mset(11, [0, 1, 2])
        # ring is now [3,0,1,2,4,5,6,7,1,2]
        self.assertEqual([0, 1, 2, 4], ring.mget(11, 4))

    # This test is skipped as it is a benchmarking only test
    def test_mperf(self):
        """
        Sanity test to make sure that mget/mset impelementations of array slicing are faster than
        basic index by index fetch/updates
        """

        def inc_mget(ring, idx, count):
            ret = []
            end_idx = idx + count
            while idx < end_idx:
                ret.append(ring.get(idx))
                idx = idx + 1
            return ret

        def inc_mset(ring, idx, elements):
            c = 0
            for e in elements:
                ring.set(idx + c, e)
                c = c + 1

        # time performance
        iterations = 200000
        set_list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        ring = RingBuffer(10)

        start = time.time()
        for t in range(iterations):
            ring.mset(t, set_list)
            self.assertEqual(set_list, ring.mget(t, 10))
        m_runtime = time.time() - start

        start = time.time()
        for t in range(iterations):
            inc_mset(ring, t, set_list)
            self.assertEqual(set_list, inc_mget(ring, t, 10))
        inc_runtime = time.time() - start

        self.logger.debug(
            'For {} iterations, mget/mset: {}s, inc_mget/inc_mset: {}s'.format(iterations, m_runtime, inc_runtime))

        self.assertTrue(m_runtime < inc_runtime)


class TestRingBufferLagStats(TestCase):
    def test_avg(self):
        s = RingBufferLagStats()
        s.sample(1)
        self.assertEqual(1, s.avg_lag)
        s.sample(3)
        self.assertEqual(2, s.avg_lag)
        s.sample(2)
        self.assertEqual(2, s.avg_lag)


class MemoryConsumer(BatchConsumer):
    def __init__(self, name, batch_size):
        super(MemoryConsumer, self).__init__(batch_size)
        self.consumed = []
        self.name = name

    def consume_batch(self, elements):
        self.consumed = self.consumed + elements

    def __str__(self):
        return str(self.name)


class BrokenConsumer(Consumer):
    exception = Exception('I am broken')

    def consume(self, elements):
        raise self.exception


class TestBatchConsumer(TestCase):
    def test_batching(self):
        c = MemoryConsumer('test', batch_size=3)
        # consume less than batch size and verify nothing happens
        c.consume([1])
        self.assertEqual([], c.consumed)
        # consume up to batch size and verify consumption occurs
        c.consume([2, 3])
        self.assertEqual([1, 2, 3], c.consumed)
        # consume over batch size and verify consumption occurs
        c.consume([4, 5, 6, 7, 8])
        self.assertEqual([1, 2, 3, 4, 5, 6], c.consumed)
        # close consumer and verify remainder of elements is consumed
        c.close()
        self.assertEqual([1, 2, 3, 4, 5, 6, 7, 8], c.consumed)


class TestDisruptor(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestDisruptor, self).__init__(*args, **kwargs)

    @classmethod
    def setUpClass(cls):
        cls.logger = logging.getLogger(cls.__name__)
        cls.logger.level = logging.DEBUG
        cls.maxDiff = None

    def random_batch_iterator(self, iterable, min_batch_size, max_batch_size):
        batch_size = random.randint(min_batch_size, max_batch_size)
        batch = []
        for e in iterable:
            batch.append(e)
            if len(batch) >= batch_size:
                yield batch
                batch_size = random.randint(min_batch_size, max_batch_size)
                batch = []
        if len(batch) > 0:
            yield batch

    def test_error_handling(self):
        handled_errors = []

        def error_handler(consumer, elements, error):
            handled_errors.append({
                'consumer': consumer,
                'elements': elements,
                'error': error
            })

        disruptor = Disruptor(1, consumer_error_handler=error_handler)
        broken_consumer = BrokenConsumer()
        working_consumer = MemoryConsumer('Working Consumer', 3)
        try:
            disruptor.register_consumer(broken_consumer)
            disruptor.register_consumer(working_consumer)
            disruptor.produce(['Test element'])
        finally:
            disruptor.close()
        # verify working consumer processed elements
        self.assertEqual(['Test element'], working_consumer.consumed)
        # verify error handler was called when broken consumer failed to consume a batch of elements
        self.assertEqual([{
            'consumer': broken_consumer,
            'elements': ['Test element'],
            'error': broken_consumer.exception
        }], handled_errors)

    def test_disruptor(self):
        elements_to_produce = 10000
        min_batch_size = 1
        max_batch_size = 12
        ring_size = 8
        n_consumers = 3

        # make some consumers with variable batch sizes
        consumers = [MemoryConsumer(i, i + 2) for i in range(n_consumers)]
        # make a disruptor
        disruptor = Disruptor(ring_size)
        try:
            # register consumers
            for consumer in consumers:
                disruptor.register_consumer(consumer)
            # make an array of elements to produce
            to_produce = [x for x in range(elements_to_produce)]
            # make a random size batched iterator of elements
            rnd_batch = self.random_batch_iterator(
                to_produce, min_batch_size=min_batch_size, max_batch_size=max_batch_size)
            # produce!
            for batch in rnd_batch:
                disruptor.produce(batch)
        finally:
            # shut down disruptor
            disruptor.close()

        # assert stats
        self.assertEqual(elements_to_produce, disruptor.stats.produced)
        self.logger.info(str(disruptor.stats))

        expected_results = to_produce
        # assert consumption
        for consumer in consumers:
            self.assertEqual(expected_results, consumer.consumed)

    def test_batch_alignment(self):
        # this verifies that batches that align perfectly do not hang disruptor.close
        consumer = MemoryConsumer('test', batch_size=1)
        # patch consumer with a latch that gets hit after it consumes any elements
        consumed_latch = Event()
        consume_original = consumer.consume

        def patched_consume(elements):
            try:
                return consume_original(elements)
            finally:
                consumed_latch.set()

        consumer.consume = patched_consume

        disruptor = Disruptor(1)
        try:
            disruptor.register_consumer(consumer)
            # produce an element
            disruptor.produce(['element'])
            # wait until element is processed by consumer
            consumed_latch.wait(10)
        finally:
            # close the disruptor to verify there is no hangup
            disruptor.close()

        self.assertEqual(['element'], consumer.consumed)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import time
from abc import abstractmethod, ABC
from typing import Any, Type

from mesh.kinds import Entity
from mesh.kinds.entity import CacheEntity
from mesh.macro import spi, mpi, T


@spi(name="mesh")
class Cache(ABC):

    @abstractmethod
    @mpi("mesh.cache.get")
    def get(self, key: str) -> CacheEntity:
        """
        :param key:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.cache.put")
    def put(self, cell: CacheEntity) -> None:
        """
        :param cell:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.cache.remove")
    def remove(self, key: str):
        """
        :param key:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.cache.incr")
    def incr(self, key: str, value: int, duration: int) -> int:
        """
        :param key:
        :param value:
        :param duration:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.cache.decr")
    def decr(self, key: str, value: int, duration: int) -> int:
        """
        :param key:
        :param value:
        :param duration:
        :return:
        """
        pass

    def get_with_type(self, key: str, kind: Type[T]) -> T:
        """
        Get cache value with generic type.
        :param key:
        :param kind:
        :return:
        """
        ce = self.get(key)
        if ce is None:
            return None
        return ce.entity().try_read_object(kind)

    def put_with_duration(self, key: str, value: Any, duration: int):
        """
        Put value to cache with expire duration.
        :param key:
        :param value:
        :param duration:
        :return:
        """
        cell = CacheEntity()
        cell.version = "1.0.0"
        cell.entity = Entity.wrap(value)
        cell.timestamp = int(time.time() * 1000)
        cell.duration = duration
        cell.key = key
        self.put(cell)

    def compute_if_absent(self, key: str, kind: Type[T], duration: int, fn):
        """
        Default compute and put if absent.
        :param key:
        :param kind:
        :param duration:
        :param fn:
        :return:
        """
        value = self.get_with_type(key, kind)
        if value is not None:
            return value
        value = fn(key)
        if value is not None:
            self.put_with_duration(key, value, duration)
        return value



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import Dict

from mesh.macro import spi, mpi


@spi("mesh")
class Cryptor(ABC):

    @abstractmethod
    @mpi("mesh.crypt.encrypt")
    def encrypt(self, buff: bytes, features: Dict[str, str]) -> bytes:
        """
        Encrypt binary to encrypted binary.
        :param buff:
        :param features:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.crypt.decrypt")
    def decrypt(self, buff: bytes, features: Dict[str, str]) -> bytes:
        """
        Decrypt binary to decrypted binary.
        :param buff:
        :param features:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.crypt.hash")
    def hash(self, buff: bytes, features: Dict[str, str]) -> bytes:
        """
        Hash compute the hash value.
        :param buff:
        :param features:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.crypt.sign")
    def hash(self, buff: bytes, features: Dict[str, str]) -> bytes:
        """
        Sign compute the signature value.
        :param buff:
        :param features:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.crypt.verify")
    def verify(self, buff: bytes, features: Dict[str, str]) -> bytes:
        """
        Verify the signature value.
        :param buff:
        :param features:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro import mps
from mesh.prsim import Endpoint, EndpointSticker


@mps
class MeshEndpoint(Endpoint, EndpointSticker[bytes, bytes]):

    def fuzzy(self, buff: bytes) -> bytes:
        pass

    def stick(self, varg: bytes) -> bytes:
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any

from mesh.ioc.paas_processor import PaaSProcessor


class MeshPostProcessor(PaaSProcessor):

    def before_initialization(self, cls: type, name: str) -> Any:
        pass

    def after_initialization(self, instance, name: str) -> Any:
        pass

    def before_instantiation(self, instance, name: str) -> Any:
        pass

    def after_instantiation(self, instance, name: str) -> Any:
        pass


#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import List

from mesh.macro import index, serializable


@serializable
class License:

    @index(0)
    def version(self) -> str:
        return ""

    @index(5)
    def level(self) -> int:
        return 0

    @index(10)
    def name(self) -> str:
        return ""

    @index(15)
    def create_by(self) -> str:
        return ""

    @index(20)
    def create_at(self) -> int:
        return 0

    @index(25)
    def active_at(self) -> int:
        return 0

    @index(30)
    def factors(self) -> List[str]:
        return []

    @index(35)
    def signature(self) -> str:
        return ""

    @index(40)
    def node_id(self) -> str:
        return ""

    @index(45)
    def inst_id(self) -> str:
        return ""

    @index(50)
    def server(self) -> str:
        return ""

    @index(55)
    def crt(self) -> str:
        return ""

    @index(60)
    def group(self) -> List[str]:
        return []

    @index(65)
    def replicas(self) -> int:
        return 0

    @index(70)
    def max_cooperators(self) -> int:
        return 0

    @index(75)
    def max_tenants(self) -> int:
        return 0

    @index(80)
    def max_users(self) -> int:
        return 0

    @index(85)
    def max_mills(self) -> int:
        return 0

    @index(90)
    def white_urns(self) -> List[str]:
        return []

    @index(95)
    def black_urns(self) -> List[str]:
        return []



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.kinds.principal import Principal
from mesh.macro import index, serializable


@serializable
class Location(Principal):

    @index(10)
    def ip(self) -> str:
        return ""

    @index(15)
    def port(self) -> str:
        return ""

    @index(20)
    def host(self) -> str:
        return ""

    @index(25)
    def name(self) -> str:
        return ""



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.http.consumer import HTTPConsumer
from mesh.http.provider import HTTPProvider

__all__ = (
    "HTTPConsumer",
    "HTTPProvider",
)


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import time
from threading import RLock, Thread, Condition


class RingBuffer(object):
    """
    Fast, list backed, preallocated Ring Buffer with static size.  Not thread safe and has no safety checking.
    """

    def __init__(self, size):
        """
        Construct a RingBuffer
        :param size: size of buffer
        :type size: int
        """
        self.buffer = []
        self.size = size
        for _ in range(self.size):
            self.buffer.append(None)

    def set(self, index, element):
        """
        Set a single element at supplied positive index; index can wrap around size of buffer
        :param index: a positive index - can wrap around size of buffer
        :type index: int
        :param element: element to set
        :type element: *
        :returns: self
        :rtype: RingBuffer
        """
        self.buffer[index % self.size] = element
        return self

    def get(self, index):
        """
        Get a single element at supplied positive index; index can wrap around size of buffer
        :param index: a positive index - can wrap around size of buffer
        :returns: element at index
        :rtype: *
        """
        return self.buffer[index % self.size]

    def mget(self, start_index, count):
        """
        Get multiple elements from buffer
        :param start_index: a positive start index - can wrap around size of buffer
        :type start_index: int
        :param count: number of elements to get - must be >= 0
        :type count: int
        :returns: list of elements
        :rtype: list
        """
        s_index = start_index % self.size
        if s_index + count > self.size:
            return self.buffer[s_index:] + self.buffer[0:(s_index + count) - self.size]
        else:
            return self.buffer[s_index:s_index + count]

    def mset(self, start_index, elements):
        """
        Set multiple elements in buffer
        :param start_index: a positive start index - can wrap around size of buffer
        :type start_index: int
        :param elements: collection of elements
        :type elements: list
        :returns: self
        :rtype: RingBuffer
        """
        s_index = start_index % self.size
        if s_index + len(elements) > self.size:
            self.buffer[s_index:self.size] = elements[0:self.size - s_index]
            self.buffer[0:len(elements) - (self.size - s_index)] = elements[self.size - s_index:len(elements)]
        else:
            self.buffer[s_index:s_index + len(elements)] = elements[0:len(elements)]


class ConsumerStats(object):
    """
    Statistics object for keeping track of consumer stats
    """

    def __init__(self, consumer_t):
        """
        Construct a consuemr for a consumer thread
        :param consumer_t: consumer thread
        :type consumer_t: ConsumerThread
        """
        self.consumer_t = consumer_t
        self.blocked_sec = 0
        self.consumed = 0
        self.consumption_sec = 0

    def report_blocked(self, sec):
        """
        Report an instance of a consumer being blocked on production for supplied seconds
        :param sec: time consumer was blocked in seconds
        :type sec: float
        """
        self.blocked_sec = self.blocked_sec + sec

    def report_consumed(self, n_elements, sec):
        """
        Report a consumer consuming a number of elements over some time period in seconds
        :param n_elements: number of elements consumed
        :type n_elements: int
        :param sec: time taken to consume elements
        :type sec: float
        """
        self.consumed = self.consumed + n_elements
        self.consumption_sec = self.consumption_sec + sec

    @property
    def cps(self):
        """
        Return elements consumed per second
        :returns: elements consumed per second
        :rtype: float
        """
        if self.consumption_sec > 0:
            return self.consumed / self.consumption_sec
        else:
            return 0

    def __str__(self):
        return '\n'.join([
            'Consumer: {}'.format(self.consumer_t.consumer),
            ' blocked_sec:{}'.format(self.blocked_sec),
            ' consumed:{}'.format(self.consumed),
            ' consume_sec:{}'.format(self.consumption_sec),
            ' cps:{}'.format(self.cps)
        ])


class RingBufferLagStats(object):
    """
    Basic disruptor ring buffer "lag" statistics
    used to keep track of how far the slowest consumer is behind producers
    """

    def __init__(self):
        self.cur_lag = 0
        self.max_lag = 0
        self.avg_lag = 0
        self.n_samples = 0

    def sample(self, lag):
        """
        Supply a lag sample
        :param lag: number of elements slowest consumer is behind producers
        :type lag: int
        """
        self.cur_lag = lag
        if lag > self.max_lag:
            self.max_lag = lag
        self.avg_lag = ((self.avg_lag * self.n_samples) + lag) / \
                       (self.n_samples + 1)
        self.n_samples = self.n_samples + 1

    def __str__(self):
        return 'cur_lag: {}, avg_lag: {}, max_lag: {}'.format(self.cur_lag, self.avg_lag, self.max_lag)


class DisruptorStats(object):
    """
    Disruptor statistics container
    """

    def __init__(self, time_fn):
        self.time_fn = time_fn
        self.consumer_stats = {}
        self.p_blocked_sec = 0
        self.produced = 0
        self.ring_lag_stats = RingBufferLagStats()
        self.start_time = time_fn()
        self.end_time = None

    def report_c_consumed(self, consumer_t, n_elements, sec):
        """
        Report a consumer thread consuming some number of elements
        :param consumer_t: a ConsumerThread
        :type consumer_t: ConsumerThread
        :param n_elements: number of elements consumed
        :type n_elements: int
        :param sec: time taken to consume elements
        :type sec: float
        """
        self.consumer_stats.setdefault(consumer_t.thread.name, ConsumerStats(
            consumer_t)).report_consumed(n_elements, sec)

    def report_p_produced(self, n_elements):
        """
        Report a producer producing some elements into the disruptor
        :param n_elements: number of elements produced
        """
        self.produced = self.produced + n_elements

    def report_c_blocked(self, consumer_t, sec):
        """
        Report an instance of a consumer being blocked on production for supplied seconds
        :param consumer_t: a ConsumerThread
        :type consumer_t: ConsumerThread
        :param sec: time consumer was blocked in seconds
        :type sec: float
        """
        self.consumer_stats.setdefault(consumer_t.thread.name, ConsumerStats(
            consumer_t)).report_blocked(sec)

    def report_p_blocked(self, sec):
        """
        Report a publisher being blocked on a "full" ring for supplied number of seconds
        :param sec: time producer was blocked in seconds
        :type sec: float
        """
        self.p_blocked_sec = self.p_blocked_sec + sec

    def report_ring_lag(self, lag_size):
        """
        Report some ring lag
        :param lag_size: number of elements slowest consumer is behind producers
        :type lag_size: int
        """
        self.ring_lag_stats.sample(lag_size)

    def close(self):
        self.end_time = self.time_fn()

    @property
    def production_sec(self):
        """
        Return number of seconds disruptor was operational
        :returns: number of seconds disruptor was operational
        :rtype: float
        """
        if self.end_time is not None:
            return self.end_time - self.start_time
        else:
            return self.time_fn() - self.start_time

    @property
    def pps(self):
        """
        Return elements produced per second
        :returns: elements produced per second
        :rtype: float
        """
        if self.production_sec > 0:
            return self.produced / self.production_sec
        else:
            return 0

    def __str__(self):
        return '\n'.join([
                             'Ring: {}'.format(self.ring_lag_stats),
                             'Producers:',
                             ' blocked_sec:{}'.format(self.p_blocked_sec),
                             ' produced:{}'.format(self.produced),
                             ' produce_sec:{}'.format(self.production_sec),
                             ' pps:{}'.format(self.pps),
                         ] + [str(value) for (key, value) in self.consumer_stats.items()])


class RingSynchronizer(object):
    """
    Disruptor synchronization uitlity
    """

    def __init__(self):
        """
        Construct RingSynchronizer
        """
        self.lock = RLock()
        self.produced_condition = Condition(self.lock)
        self.consumed_condition = Condition(self.lock)

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, t, v, tb):
        self.lock.release()

    def await_production(self, timeout_sec=5):
        """
        Block calling thread for supplied amount of time or until a production condition occurs
        :param timeout_sec: timeout seconds
        :type timeout_sec: float
        """
        with self.lock:
            self.produced_condition.wait(timeout_sec)

    def await_consumption(self, timeout_sec=5):
        """
        Block calling thread for supplied amount of time or until a consumption condition occurs
        :param timeout_sec: timeout seconds
        :type timeout_sec: float
        """
        with self.lock:
            self.consumed_condition.wait(timeout_sec)

    def notify_production(self):
        """
        Notify and unblock all threads waiting on a production condition
        """
        with self.lock:
            self.produced_condition.notify_all()

    def notify_consumption(self):
        """
        Notify and unblock all threads waiting on a consumption condition
        """
        with self.lock:
            self.consumed_condition.notify_all()


class ConsumerThread(object):
    """
    A disruptor consumption thread wrapper
    """

    def __init__(self, disruptor, consumer):
        self.disruptor = disruptor
        self.consumer = consumer
        self.seqnum = disruptor.producer_seqnum
        self.thread = Thread(
            name='{}_Consumer_{}'.format(disruptor.name, consumer), target=self.run)
        self.thread.start()

    def run(self):
        """
        Consume while there's something to consume!
        """
        available_count = 0
        to_consume = None
        while self.disruptor.running:

            # report ring lag once in a while
            self.disruptor._unsafe_report_lag()

            while available_count == 0 and self.disruptor.running:
                # atomically check+fetch available data and go to sleep if none
                with self.disruptor.sync:
                    available_count = self.disruptor.producer_seqnum - self.seqnum
                    if available_count > 0:
                        to_consume = self.disruptor.ring_buffer.mget(
                            self.seqnum, available_count)
                    else:
                        s = self.disruptor.time_fn()
                        self.disruptor.sync.await_production()
                        self.disruptor.stats.report_c_blocked(
                            self, self.disruptor.time_fn() - s)

            # when data is available, consume it *outside* of lock!
            self._consume_safe(to_consume)

            # re-lock after consumption to update state
            with self.disruptor.sync:
                self.seqnum = self.seqnum + available_count
                self.disruptor.sync.notify_consumption()
                available_count = 0
                to_consume = None

        # after disruptor stops, consume the rest of available data
        to_consume = None
        with self.disruptor.sync:
            available_count = self.disruptor.producer_seqnum - self.seqnum
            if available_count > 0:
                to_consume = self.disruptor.ring_buffer.mget(
                    self.seqnum, available_count)
        self._consume_safe(to_consume)
        self.consumer.close()

    def _consume_safe(self, elements):
        if elements is not None and len(elements) > 0:
            s = self.disruptor.time_fn()
            try:
                self.consumer.consume(elements)
            except Exception as e:
                if self.disruptor.consumer_error_handler is not None:
                    self.disruptor.consumer_error_handler(
                        self.consumer, elements, e)
            self.disruptor.stats.report_c_consumed(
                self, len(elements), self.disruptor.time_fn() - s)


class Disruptor(object):
    """
    A basic multi-threaded RingBuffer; a Disruptor-Lite implementation using a thread-per-consumer model.
    This is an efficient alternative to queues in cases where multiple workloads need to consume the same type of data.
    See the below links for an explanation of how this works:
        * http://mechanitis.blogspot.com/2011/07/dissecting-disruptor-writing-to-ring.html explanation of structure
        * https://lmax-exchange.github.io/disruptor/ original Disruptor
        * https://medium.com/@teivah/understanding-the-lmax-disruptor-caaaa2721496 - mechanical sympathy (largely not applicable in python)
    Allows for many producers to efficiently concurrently produce elements to be concurrently consumed by many consumers.
    Uses a shared ring buffer between consumers and producers with a minimum amount of blocking.  An element in the ring
    can only be written to after all consumers have consumed it, effectively handling backpressure from any number of
    parallel consumers.
    Producers (callers of the produce method) are only blocked when the ring is full.
    Unlike a C/C++/C#/Java disruptor, this implementation uses Python threads, so it offers little to no mechanical sympathy
    outside of supporting lock-free parallel consumptions of big sections of the ring buffer which may or may not fall into
    a single CPU cache line.  Even so, this is a great tool for processing multiple events of the same type concurrently.
    """

    def __init__(self, size=1024, name='Disruptor', consumer_error_handler=None, time_fn=time.time):
        """
        Initialize a disruptor of supplied size
        :param size: size of disruptor ring buffer
        :type size: int
        :param name: disruptor name (used in naming consumer threads)
        :type name: str
        :param consumer_error_handler: optional function invoked if a consumer fails to consume data.  Must accept a consumer instance, input to consumer, and error
        :type consumer_error_handler: Function
        :param time_fn: zero argument time provider function - defaults to time.time(); used strictly for statistics
        :type time_fn: Function
        """
        self.name = name
        self.time_fn = time_fn
        self.consumer_error_handler = consumer_error_handler
        self.stats = DisruptorStats(time_fn)
        self.ring_buffer = RingBuffer(size)
        self.sync = RingSynchronizer()
        self.producer_seqnum = 0
        self.consumers = []
        self.running = True

    def produce(self, elements):
        """
        Produce multiple elements, blocking if the disruptor is full
        :param elements: list of elements to produce
        :type elements: list
        """
        if not self.running:
            raise Exception('Disruptor is stopped')

        # report ring lag once in a while
        self._unsafe_report_lag()

        produced = 0

        while produced < len(elements):
            with self.sync.lock:
                # figure out maximum number of elements producer can produce
                # the producer can produce up to the slowest consumer slot
                #
                # Example:
                #
                # - Ring size = 30
                #
                #                     seq#  ring idx
                # - Consumer 1 (C1) = 83    13
                # - Consumer 2 (C2) = 78    8
                # - Producer seq    = 88    18
                #
                #                  C2   C1   P
                #                  |    |    |
                # ringidx  00   05 | 10 | 15 | 20   30
                #          |----|--|-|--|-|--|-|----|
                # seqnum   70   75 | 80 | 85 | 90   100
                #                  |    |    |======>
                #          =======>|    |    |
                #
                # The producer can produce up to the slowest consumer's (C2) slot
                # in the ring (8) - outlined by "==>"
                #
                # There's several ways of computing this:
                #
                # * (ring_size - p_idx) + min(C1_idx, C2_idx)
                #   (30        - 18   ) + 8                  = 20
                # * ring_size - p_seq + min(C1_seq, C2_seq)
                #   30        - 88    + 78                   = 20
                #
                # This code uses the second approach because it doesn't require
                # computing an index in the ring buffer.
                can_produce = self.ring_buffer.size - self.producer_seqnum + \
                              min(map(lambda c: c.seqnum, self.consumers))
                if can_produce <= 0:
                    s = self.time_fn()
                    self.sync.await_consumption()
                    self.stats.report_p_blocked(self.time_fn() - s)
                else:
                    to_produce_cnt = min(can_produce, len(elements) - produced)
                    self.ring_buffer.mset(
                        self.producer_seqnum, elements[produced:produced + to_produce_cnt:])
                    produced += to_produce_cnt
                    self.producer_seqnum += to_produce_cnt
                    self.sync.notify_production()

        self.stats.report_p_produced(produced)

    def close(self):
        """
        Close the disruptor, block (potentially indefinitely) until consumers finish consuming all elements
        """
        with self.sync.lock:
            if self.running:
                self.running = False
                # wake up any threads waiting on production
                self.sync.notify_production()
        for consumer in self.consumers:
            consumer.thread.join()
        self.stats.close()

    def register_consumer(self, consumer):
        """
        Register a consumer for this disruptor.  Once registered, a consumer can not be de-registered!
        :param consumer: a Consumer
        :type consumer: Consumer
        """
        self.consumers.append(ConsumerThread(self, consumer))
        return self

    def _unsafe_report_lag(self):
        if self.producer_seqnum % 10 == 3:
            self.stats.report_ring_lag(
                self.producer_seqnum - min(map(lambda c: c.seqnum, self.consumers)))

    def __str__(self):
        return 'Disruptor-"{}"'.format(self.name)

    def __del__(self):
        self.close()


class Consumer(object):
    """
    A no-op consumer class; can be extended for convenience
    """

    def consume(self, elements):
        """
        Consume elements
        :param elements: a list of elements
        :type elements: list
        """
        raise NotImplementedError()

    def close(self):
        """
        Close consumer, finishing any consumption tasks e.g. terminating batches
        """
        pass

    def __str__(self):
        return self.__class__.__name__


class BatchConsumer(Consumer):
    """
    A Consumer that buffers incoming elements into batches of specified size without using iteration.
    Useful for workloads that are optimized to a specific batch size.
    """

    def __init__(self, batch_size):
        """
        Construct a BatchConsumer with supplied batch size
        :param batch_size: a batch size
        :type batch_size: int
        """
        self.batch_size = batch_size
        self._batch = []

    def consume(self, elements):
        consumed = 0
        while consumed < len(elements):
            to_consume_cnt = min(
                self.batch_size - len(self._batch), len(elements) - consumed)
            self._batch.extend(elements[consumed:consumed + to_consume_cnt])
            if len(self._batch) >= self.batch_size:
                self.consume_batch(self._batch)
                self._batch = []
            consumed += to_consume_cnt

    def close(self):
        if len(self._batch) > 0:
            self.consume_batch(self._batch)

    def consume_batch(self, batch):
        """
        Consume a batch of elements
        :param batch: a batch of elements of batch_size or smaller
        :type batch: list
        """
        raise NotImplementedError()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod
from typing import List, Any

from mesh.kinds.event import Event
from mesh.kinds.event import Topic
from mesh.kinds.principal import Principal
from mesh.macro import mpi, spi


@spi("mesh")
class Publisher(ABC):

    @abstractmethod
    @mpi("mesh.queue.publish")
    def publish(self, events: List[Event]) -> List[str]:
        """
        Publish event to mesh.
        :param events:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.queue.broadcast")
    def broadcast(self, events: List[Event]) -> List[str]:
        """
        Synchronized broadcast the event to all subscriber. This maybe timeout with to many subscriber.
        :param events: Event payload
        :return: Synchronized subscriber return value
        """
        pass

    def publish_with_topic(self, binding: Topic, payload: Any) -> str:
        """
        Publish message to local node.
        :param binding:
        :param payload:
        :return:
        """
        event = Event.new_instance(payload, binding)
        return self.publish([event])[0]

    def unicast(self, binding: Topic, payload: Any, principal: Principal) -> str:
        """
        Unicast will publish to another node.
        :param binding:
        :param payload:
        :param principal:
        :return:
        """
        event = Event.new_instance_with_target(payload, binding, principal)
        return self.publish([event])[0]

    def multicast(self, binding: Topic, payload: Any, principals: List[Principal]) -> List[str]:
        """
        Multicast will publish event to principal groups.
        :param binding:
        :param payload:
        :param principals:
        :return:
        """
        events = []
        for principal in principals:
            events.append(Event.new_instance_with_target(payload, binding, principal))
        return self.publish(events)

    def broadcast_with_topic(self, binding: Topic, payload: Any) -> List[str]:
        """
        Synchronized broadcast the event to all subscriber. This maybe timeout with to many subscriber.
        :param binding:
        :param payload:
        :return:
        """
        return self.broadcast([Event.new_instance(payload, binding)])



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import inspect
import os
from typing import Type, Any, Generic, Dict, List, Callable

import mesh.kinds as kinds
import mesh.log as log
import mesh.tool as tool
from mesh.kinds import Reference, Service, Environ, MeshFlag, Event, Registration, Resource, Topic, Versions
from mesh.macro import ark, T, A, mpi, MPI, spi, mps, MPS, binding, Binding, ServiceLoader, Returns, Parameters, \
    Inspector
from mesh.mpc.compiler import Compiler
from mesh.mpc.eden import Eden
from mesh.mpc.inspector import MethodInspector
from mesh.mpc.invoker import Invocation, Execution
from mesh.mpc.service import ServiceInvokeHandler
from mesh.mpc.service_proxy import ServiceProxy
from mesh.mpc.urn import URN, URNFlag, MESH_DOMAIN
from mesh.prsim import Network, RuntimeHook, Subscriber, Registry, Scheduler, Builtin


class Provider:

    def __init__(self, kind: Type[T], service: Any):
        self.making: bool = False
        self.kind: Type[T] = kind
        self.service: Any = service


class Consumer:
    def __init__(self, metadata: mpi, reference: Type[T]):
        self.metadata: mpi = metadata
        self.reference: Type[T] = reference
        self.proxy: Any = ServiceProxy.static_proxy(reference, metadata)


class Instance(Generic[T], Execution[T]):

    def __init__(self, urn: str, kind: Type[T], method: Any, target: Any, resource: T):
        declared_method = getattr(kind, method.__name__)
        self.urn = urn
        self.kind = kind
        self.method = method
        self.target = target
        self.resource = resource
        self.inspector = MethodInspector(MPI.get_mpi_if_present(method), kind, method)
        self.invoker = ServiceInvokeHandler(target)
        self.__intype = ServiceLoader.load(Compiler).get_default().intype(declared_method)
        self.__retype = ServiceLoader.load(Compiler).get_default().retype(declared_method)

    def schema(self) -> T:
        return self.resource

    def inspect(self) -> Inspector:
        return self.inspector

    def intype(self) -> Type[Parameters]:
        return self.__intype

    def retype(self) -> Type[Returns]:
        return self.__retype

    def inflect(self) -> Parameters:
        return self.__intype()

    def reflect(self) -> Returns:
        return self.__retype()

    def run(self, invocation: Invocation) -> Any:
        return self.invoker.run(invocation)


@spi("mesh")
@binding(topic="mesh.registry.event", code="refresh", meshable=False)
class MeshEden(Eden, RuntimeHook, Subscriber):
    """
    Like: create.tenant.omega.json.http2.lx000001.mpi.trustbe.net
    """

    def __init__(self):
        self.topic = Topic()
        self.topic.topic = "mesh.registry.event"
        self.topic.code = "refresh"
        self.task_id = ""
        self.providers: Dict[Type[T], Provider] = {}
        self.consumers: Dict[Type[T], Dict[mpi, Consumer]] = {}
        self.indies: Dict[Type[T], Dict[mpi, Dict[Any, Instance[Reference]]]] = {}
        self.services: Dict[str, Instance[Service]] = {}
        self.references: Dict[str, Instance[Reference]] = {}

    def define(self, metadata: mpi, reference: Type[T]) -> T:
        return self.make_consumer(metadata, reference)

    def refer(self, metadata: mpi, reference: Type[T], method: Inspector) -> Execution[Reference]:
        return self.make_reference_exec(metadata, reference, self.get_environ()).get(method.get_signature())

    def store(self, kind: Type[T], service: Any):
        if not self.providers.get(kind):
            self.providers[kind] = Provider(kind, service)
        return self.providers.get(kind)

    def infer(self, urn: str) -> Execution[Service]:
        return self.make_service_exec(self.get_environ()).get(URN.parse(urn).name)

    def refer_types(self) -> List[Type[T]]:
        return list(self.consumers.keys())

    def infer_types(self) -> List[Type[T]]:
        types: List[Type[T]] = []
        for provider in self.providers.values():
            types.append(provider.kind)
        return types

    def make_consumer(self, metadata: mpi, reference: Type[T]):
        if not self.consumers.get(reference):
            self.consumers[reference] = {}
        if not self.consumers.get(reference).get(metadata):
            self.consumers[reference][metadata] = Consumer(metadata, reference)
        return self.consumers.get(reference).get(metadata)

    def make_reference_exec(self, md: mpi, ref: Type[T], env: Environ) -> Dict[Any, Instance[Reference]]:
        consumer = self.make_consumer(md, ref)
        if not self.indies.get(ref):
            self.indies[ref] = {}
        if not self.indies.get(ref).get(md):
            executions: Dict[Any, Instance[Reference]] = {}
            for kind, methods in tool.get_declared_methods(ref).items():
                for method in methods:
                    refer = self.make_method_as_reference(env, md, kind, method)
                    instance = Instance(refer.urn, ref, method, consumer.proxy, refer)
                    self.references[refer.urn] = instance
                    executions[instance.inspect().get_signature()] = instance
            self.indies.get(ref)[md] = executions
        return self.indies[ref][md]

    def make_service_exec(self, environ: Environ) -> Dict[str, Instance[Service]]:
        for provider in self.providers.values():
            if provider.making:
                continue
            for kind, methods in tool.get_declared_methods(provider.kind).items():
                for method in methods:
                    if not hasattr(provider.service, method.__name__):
                        continue
                    mm = getattr(provider.service, method.__name__)
                    schemas: List[Service] = self.make_schema(environ, provider.kind, kind, method)
                    for schema in schemas:
                        urn = URN.parse(schema.urn)
                        self.services[urn.name] = Instance(schema.urn, provider.kind, mm, provider.service, schema)
            provider.making = True
        return self.services

    @staticmethod
    def get_environ() -> Environ:
        network = ServiceLoader.load(Network).get_default()
        return network.get_environ()

    def make_schema(self, environ: Environ, interface: T, kind: T, method: Any) -> List[Service]:
        schemas = []
        macro = MPS.get_mps_if_present(interface)
        if macro:
            schemas.append(self.make_mps_as_service(macro, kind, method, environ))
        bindings = Binding.get_binding_if_present(interface)
        for metadata in bindings:
            if metadata.meshable:
                schemas = schemas.append(self.make_binding_as_service(metadata, kind, method, environ))
        return schemas

    def make_method_as_reference(self, environ: Environ, metadata: MPI, interface: T, method: Any) -> Reference:
        om = MPI.get_mpi_if_present(method)
        ok = MPI.get_mpi_if_present(interface)
        name = self.choose_if_present(lambda x: x.name, "", om)
        alias = tool.anyone(name, str.lower(f'{interface.__name__}.{method.__name__}'))
        reference = Reference()
        reference.namespace = interface.__name__
        reference.name = method.__name__
        reference.version = self.choose_if_present(lambda x: x.version, '', metadata, om, ok)
        reference.proto = self.choose_if_present(lambda x: x.proto, MeshFlag.GRPC.name, metadata, om, ok)
        reference.codec = self.choose_if_present(lambda x: x.codec, MeshFlag.JSON.name, metadata, om, ok)
        reference.flags = self.choose_if_present(lambda x: x.flags, 0, om, ok) | metadata.flags
        reference.timeout = self.choose(lambda x: x.timeout, 10000, metadata, om, ok)
        reference.retries = self.choose(lambda x: x.retries, 3, metadata, om, ok)
        reference.node = self.choose_if_present(lambda x: x.node, "", metadata, om, ok)
        reference.inst = self.choose_if_present(lambda x: x.inst, "", metadata, om, ok)
        reference.zone = self.choose_if_present(lambda x: x.zone, "", metadata, om, ok)
        reference.cluster = self.choose_if_present(lambda x: x.cluster, "", metadata, om, ok)
        reference.cell = self.choose_if_present(lambda x: x.cell, "", metadata, om, ok)
        reference.group = self.choose_if_present(lambda x: x.group, "", metadata, om, ok)
        reference.address = self.choose_if_present(lambda x: x.address, "", metadata, om, ok)
        reference.urn = self.get_urn(alias, self.get_reference_urn_flag(reference), environ)
        return reference

    def make_mps_as_service(self, metadata: MPS, kind: T, method: Any, environ: Environ) -> Service:
        om = MPI.get_mpi_if_present(method)
        ok = MPI.get_mpi_if_present(kind)
        name = self.choose_if_present(lambda x: x.name, "", om).replace("${mesh.name}", tool.get_mesh_name()). \
            replace("${mesh.uname}", metadata.name)
        alias = tool.anyone(name, str.lower(f'{kind.__name__}.{method.__name__}'))
        service = self.init_service(environ)
        service.kind = kinds.MPS
        service.namespace = kind.__name__
        service.name = method.__name__
        service.proto = self.choose_if_present(lambda x: x.proto, MeshFlag.GRPC.name, metadata, om, ok)
        service.codec = self.choose_if_present(lambda x: x.codec, MeshFlag.JSON.name, metadata, om, ok)
        service.version = self.choose_if_present(lambda x: x.version, metadata.version, metadata, om, ok)
        service.flags = self.choose_if_present(lambda x: x.flags, 0, om, ok) | metadata.flags
        service.timeout = self.choose(lambda x: x.timeout, metadata.timeout, metadata, om, ok)
        service.urn = self.get_urn(alias, self.get_service_urn_flag(service), environ)
        return service

    def make_binding_as_service(self, metadata: Binding, kind: T, method: Any, environ: Environ) -> Service:
        service = self.init_service(environ)
        service.kind = kinds.Binding
        service.namespace = metadata.topic
        service.name = metadata.code
        service.proto = self.choose_if_present(lambda x: x.proto, MeshFlag.GRPC.name, metadata)
        service.codec = self.choose_if_present(lambda x: x.codec, MeshFlag.JSON.name, metadata)
        service.version = metadata.version
        service.flags = metadata.flags
        service.timeout = self.choose(lambda x: x.timeout, metadata.timeout, metadata)
        service.urn = self.get_urn(str.lower(f'{metadata.topic}.{metadata.code}'), self.get_service_urn_flag(service),
                                   environ)
        return service

    def init_service(self, environ: Environ) -> Service:
        service = Service()
        service.namespace = ''
        service.name = ''
        service.version = ''
        service.proto = ''
        service.codec = ''
        service.flags = 0
        service.timeout = 10000
        service.retries = 0
        service.node = environ.node_id
        service.inst = environ.inst_id
        service.zone = environ.lattice.zone
        service.cluster = environ.lattice.cluster
        service.cell = environ.lattice.cell
        service.group = environ.lattice.group
        service.address = self.get_listened_address()
        service.kind = kinds.MPS
        service.lang = 'Python3'
        service.attrs = {}
        service.urn = ''
        return service

    @staticmethod
    def get_urn(alias: str, flag: URNFlag, environ: Environ) -> str:
        urn = URN()
        urn.domain = MESH_DOMAIN
        urn.node_id = str.lower(environ.node_id)
        urn.name = alias
        urn.flag = flag
        return urn.string()

    @staticmethod
    def get_reference_urn_flag(reference: Reference) -> URNFlag:
        authority = str.split(f'{reference.address}:', ':')
        flag = URNFlag()
        return flag.reset(MeshFlag.of_name(reference.proto).get_code(),
                          MeshFlag.of_name(reference.codec).get_code(),
                          reference.version,
                          reference.zone,
                          reference.cluster,
                          reference.cell,
                          reference.group,
                          authority[0],
                          authority[1])

    @staticmethod
    def get_service_urn_flag(service: Service) -> URNFlag:
        flag = URNFlag()
        return flag.reset(MeshFlag.of_name(service.proto).get_code(),
                          MeshFlag.of_name(service.codec).get_code(),
                          service.version,
                          service.zone,
                          service.cluster,
                          service.cell,
                          service.group,
                          '',
                          '')

    @staticmethod
    def choose_if_present(fn: Callable[[A], T], defaults: T, *tags: A) -> T:
        for tag in tags:
            if tag and fn(tag):
                return fn(tag)

        return defaults

    @staticmethod
    def choose(fn: Callable[[A], T], defaults: T, *tags: A) -> T:
        for tag in tags:
            if tag and fn(tag) and fn(tag) != defaults:
                return fn(tag)

        return defaults

    @staticmethod
    def get_listened_address() -> str:
        return f'{tool.get_mesh_runtime().hostname}:{tool.get_mesh_runtime().port}'

    def new_registration(self, metadata: Resource, attachments: Dict[str, str]) -> Registration:
        registration = Registration()
        registration.instance_id = self.get_listened_address()
        registration.content = metadata
        registration.kind = Registration.METADATA
        registration.address = self.get_listened_address()
        registration.name = tool.get_mesh_name()
        registration.timestamp = 30 * 1000
        registration.attachments = attachments
        return registration

    @mpi
    def registry(self) -> Registry:
        """ """
        pass

    @spi("mesh")
    def scheduler(self) -> Scheduler:
        return ServiceLoader.load(Scheduler).get_default()

    @spi("mesh")
    def builtin(self) -> Builtin:
        return ServiceLoader.load(Builtin).get_default()

    def start(self):
        for kind, nss in ark.export(mps).items():
            for spec in ark.providers(mps, kind):
                if not isinstance(spec.metadata, MPS) or inspect.isabstract(spec.kind):
                    continue
                try:
                    self.store(kind, spec.kind())
                    break
                except BaseException as e:
                    log.warn(f"Mesh service {kind.__name__} with {e.__str__()}")

        if os.environ.get("MESH_REGISTER", "0") != "1":
            return

        self.task_id = self.scheduler().period(30 * 1000, self.topic)

    def stop(self):
        self.scheduler().cancel(self.task_id)
        self.scheduler().shutdown(3 * 1000)
        metadata = Resource()
        metadata.services = []
        metadata.references = []
        try:
            self.registry().unregister(self.new_registration(metadata, {}))
        except BaseException as e:
            log.warn("Graceful unregister metadata, {}", str(e))

    def refresh(self):
        self.start()

    def subscribe(self, event: Event):
        v = self.builtin().version()
        version = v if v else Versions()
        attachments = {}
        for k, v in (version.infos if version.infos else {}).items():
            attachments[k] = v
        environ = self.get_environ()
        self.make_service_exec(environ)
        for _, consumers in self.consumers.items():
            for _, consumer in consumers.items():
                self.make_reference_exec(consumer.metadata, consumer.reference, environ)
        metadata = Resource()
        metadata.services = []
        for _, reference in self.references.items():
            metadata.services.append(reference.resource)
        metadata.references = []
        for _, service in self.services.items():
            metadata.references.append(service.resource)
        self.registry().register(self.new_registration(metadata, attachments))



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any, List, Dict

from mesh.kinds.reference import Reference
from mesh.kinds.service import Service
from mesh.macro import index, serializable

Binding = "Binding"
MPS = "MPS"
Forward = "Forward"
Proxy = "Proxy"


@serializable
class Registration:
    METADATA = "metadata"
    PROXY = "proxy"
    SERVER = "server"
    COMPLEX = "complex"

    @index(0)
    def instance_id(self) -> str:
        return ''

    @index(5)
    def name(self) -> str:
        return ''

    @index(10)
    def kind(self) -> str:
        return ''

    @index(15)
    def address(self) -> str:
        return ''

    @index(20)
    def content(self) -> Any:
        return None

    @index(25)
    def timestamp(self) -> int:
        return 0

    @index(30)
    def attachments(self) -> Dict[str, str]:
        return {}


class Resource:

    @index(0)
    def references(self) -> List[Reference]:
        return []

    @index(5)
    def services(self) -> List[Service]:
        return []



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import inspect
from abc import abstractmethod, ABC
from typing import Any, Type, Dict, List

import mesh.tool as tool
from mesh.macro import Parameters, Returns, spi, Index, Cause, serializable

ATTACHMENTS: str = "attachments"
CODE: str = "code"
MESSAGE: str = "message"
CONTENT: str = "content"
CAUSE: str = "cause"


@spi(name="mpyc")
class Compiler(ABC):

    @abstractmethod
    def intype(self, method: Any) -> Type[Parameters]:
        """
        Get the parameter class type.
        :param method:
        :return:
        """
        pass

    @abstractmethod
    def retype(self, method: Any) -> Type[Returns]:
        """
        Get the return class type.
        :param method:
        :return:
        """
        pass

    @abstractmethod
    def compile(self, code: str, name: str, add_to_sys_modules=0):
        """
        Compile source code.
        :param code:
        :param name:
        :param add_to_sys_modules:
        :return:
        """
        pass


@spi(name="mpyc")
class PythonCompiler(Compiler):
    cache = {}

    @staticmethod
    def make_getter(name, kind: type):
        def getter(self) -> kind:
            if hasattr(self, name):
                return getattr(self, name)
            return None

        return getter

    @staticmethod
    def make_setter(name):
        def setter(self, value):
            setattr(self, name, value)

        return setter

    @staticmethod
    def make_kind():
        def kind(self) -> type:
            return self.__cls__

        return kind

    @staticmethod
    def make_map():
        def maps(self, names: List[str]) -> Dict[str, Any]:
            values = {}
            for name in names:
                if hasattr(self, name):
                    values[name] = getattr(self, name)
            return values

        return maps

    @staticmethod
    def make_get_arguments(names: List[str]):
        def getter(self) -> List[Any]:
            values = []
            for name in names:
                if hasattr(self, name):
                    values.append(getattr(self, name))
                else:
                    values.append(None)
            return values

        return getter

    @staticmethod
    def make_set_arguments(names: List[str]):
        def setter(self, *arguments):
            if not arguments:
                return
            for idx, name in enumerate(names):
                if idx < arguments.__len__():
                    setattr(self, name, arguments[idx])

        return setter

    @staticmethod
    def make_decode(names: List[str]):
        def setter(self, values: List[Any]):
            if not values:
                return
            for idx, name in enumerate(names):
                if idx < values.__len__():
                    setattr(self, name, values[idx])

        return setter

    @staticmethod
    def make_class_name(method: Any, suffix: str) -> str:
        qualname = tool.split(method.__qualname__, ".")
        return f"Mesh{qualname[0]}{qualname[1].title()}{suffix}".replace("_", "")

    def intype(self, method: Any) -> Type[Parameters]:
        signature = inspect.signature(method)
        class_name = self.make_class_name(method, "Parameters")
        intype = self.cache.get(class_name, None)
        if intype:
            return intype
        names = []
        annotations = {ATTACHMENTS: Dict[str, str]}
        for idx, (name, parameter) in enumerate(signature.parameters.items()):
            if idx > 0:
                names.append(name)
                annotations[name] = parameter.annotation
        variables = {
            "get_attachments": self.make_getter(ATTACHMENTS, Dict[str, str]),
            "set_attachments": self.make_setter(ATTACHMENTS),
            "map": self.make_map(),
            "kind": self.make_kind(),
            "get_arguments": self.make_get_arguments(names),
            "set_arguments": self.make_set_arguments(names),
            "__annotations__": annotations,
        }
        for idx, (name, parameter) in enumerate(signature.parameters.items()):
            if idx == 0:
                variables[ATTACHMENTS] = Index(idx=-1, name=ATTACHMENTS, kind=Dict)
                continue
            variables[name] = Index(idx=idx, name=name, kind=parameter.annotation)
        intype = serializable(type(class_name, (Parameters,), variables))
        self.cache.__setitem__(class_name, intype)
        return intype

    def retype(self, method: Any) -> Type[Returns]:
        signature = inspect.signature(method)
        class_name = self.make_class_name(method, "Returns")
        retype = self.cache.get(class_name, None)
        if retype:
            return retype
        variables = {
            CODE: Index(idx=0, name=CODE, kind=str),
            MESSAGE: Index(idx=5, name=MESSAGE, kind=str),
            CAUSE: Index(idx=10, name=CAUSE, kind=Cause),
            CONTENT: Index(idx=15, name=CONTENT, kind=signature.return_annotation),
            "get_code": self.make_getter(CODE, str),
            "set_code": self.make_setter(CODE),
            "get_message": self.make_getter(MESSAGE, str),
            "set_message": self.make_setter(MESSAGE),
            "get_cause": self.make_getter(CAUSE, Cause),
            "set_cause": self.make_setter(CAUSE),
            "get_content": self.make_getter(CONTENT, signature.return_annotation),
            "set_content": self.make_setter(CONTENT),
            "__annotations__": {
                CODE: str,
                MESSAGE: str,
                CAUSE: Cause,
                CONTENT: signature.return_annotation
            },
        }
        retype = serializable(type(class_name, (Returns,), variables))
        self.cache.__setitem__(class_name, retype)
        return retype

    def compile(self, code: str, name: str, add_to_sys_modules=0):
        """
        Import dynamically generated code as a module. code is the
        object containing the code (a string, a file handle or an
        actual compiled code object, same types as accepted by an
        exec statement). The name is the name to give to the module,
        and the final argument says whether to add it to sys.modules
        or not. If it is added, a subsequent import statement using
        name will return this module. If it is not added to sys.modules
        import will try to load it in the normal fashion.
        """
        import sys
        import types
        module = types.ModuleType(name)
        exec(code, module.__dict__)
        if add_to_sys_modules:
            sys.modules[name] = module
        return module



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import abstractmethod, ABC
from typing import Type, Any, get_type_hints, List

import mesh.codec.tools as tools
from mesh.macro import Index, spi, T, ServiceLoader


class Context:

    def __init__(self):
        self.offset = 0

    def form(self, value: Any, cls: Type[T]) -> T:
        """ Transform given value. """
        self.offset = self.offset + 1

    def inform(self, value: Any, cls: Type[T]) -> Any:
        self.offset = self.offset + 1

    @staticmethod
    def var_type(cls: Type[T], name: str):
        ft = get_type_hints(cls).get(name, None)
        if ft:
            return ft
        cls = getattr(cls, '__dict__', {}).get(name, None)
        return bytes if cls is Index and cls.kind is bytes else str


class ChainContext(Context):

    def __init__(self, ctx: Context):
        self.ctx = ctx

    def form(self, value: Any, cls: Type[T]) -> T:
        return super().form(value, cls)

    def inform(self, value: Any, cls: Type[T]) -> Any:
        return super().inform(value, cls)


@spi("mesh")
class Former(ABC):

    @abstractmethod
    def form(self, ctx: Context, value: Any, cls: Type[T]) -> T:
        """ Transform given value. """
        pass

    @abstractmethod
    def inform(self, ctx: Context, value: Any, cls: Type[T]) -> Any:
        """ Transform from given value. """
        pass


@spi("object")
class ObjectFormer(Former):

    def form(self, ctx: Context, value: Any, cls: Type[T]) -> T:
        if cls is not dict:
            return ctx.form(value, cls)

    def inform(self, ctx: Context, value: Any, cls: Type[T]) -> Any:
        return self.restore(value, cls)


@spi("bytes")
class BytesFormer(Former):

    def form(self, ctx: Context, value: Any, cls: Type[T]) -> T:
        if type(value) is bytes and (cls is bytes or (cls is Index and cls.kind is bytes)):
            return tools.b64encode(value).encode('utf-8')
        return ctx.form(value, cls)

    def inform(self, ctx: Context, value: Any, cls: Type[T]) -> Any:
        if type(value) is str and (cls is bytes or (cls is Index and cls.kind is bytes)):
            return tools.b64decode(value.encode('utf-8'))
        return ctx.inform(value, cls)


@spi("list")
class ListFormer(Former):

    def form(self, ctx: Context, value: Any, cls: Type[T]) -> T:
        if cls is not list:
            return ctx.form(value, cls)
        pass

    def inform(self, ctx: Context, value: Any, cls: Type[T]) -> Any:
        if cls is not list:
            return ctx.form(value, cls)


class PuppetFormer(Former):

    def __init__(self, informers: List[Former]):
        self.informers = informers

    def form(self, ctx: Context, value: Any, cls: Type[T]) -> T:
        if ctx.offset >= self.informers.__len__():
            return value
        for (index, informer) in self.informers:
            v = informer.form(ctx, value, cls)
            if index == ctx.offset:
                return v
        return value

    def inform(self, ctx: Context, value: Any, cls: Type[T]) -> Any:
        if ctx.offset >= self.informers.__len__():
            return value
        for (index, informer) in self.informers:
            v = informer.inform(ctx, value, cls)
            if index == ctx.offset:
                return v
        return value


@spi("mesh")
class MeshFormer(Former):

    def get_once(self) -> Former:
        if not hasattr(self, '__former__'):
            setattr(self, '__former__', PuppetFormer(ServiceLoader.load(Former).list('')))
        return self.__transformer__

    def matches(self, ctx: Context, value: Any, cls: Type[T]) -> bool:
        return self.get_once().matches(ctx, value, cls)

    def form(self, ctx: Context, value: Any, cls: Type[T]) -> T:
        return self.get_once().form(ctx, value, cls)

    def inform(self, ctx: Context, value: Any, cls: Type[T]) -> Any:
        return self.get_once().inform(ctx, value, cls)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any

import mesh.log as log
from mesh.macro import spi
from mesh.mpc import Provider


@spi("http")
class HTTPProvider(Provider):

    def start(self, address: str, tc: Any):
        log.info(f"Listening and serving HTTP 1.x on {address}")

    def close(self):
        log.info(f"Graceful stop HTTP 1.x serving")



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any, List

from mesh.kinds import Document, Paging, Page
from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import DataHouse


@spi("mesh")
class MeshDataHouse(DataHouse):

    def __init__(self):
        self.proxy = ServiceProxy.default_proxy(DataHouse)

    def writes(self, docs: List[Document]):
        return self.proxy.writes(docs)

    def write(self, doc: Document):
        return self.proxy.write(doc)

    def read(self, index: Paging) -> Page[Any]:
        return self.proxy.read(index)

    def indies(self, index: Paging) -> Page[Any]:
        return self.proxy.indies(index)

    def tables(self, index: Paging) -> Page[Any]:
        return self.proxy.tables(index)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any, Dict

import mesh.log as log
from mesh.cause import MeshCode, Codeable, MeshException
from mesh.codec import Codec
from mesh.context import Mesh
from mesh.kinds import MeshFlag
from mesh.macro import spi, ServiceLoader
from mesh.mpc.digest import Digest
from mesh.mpc.filter import Filter, Invoker, Invocation, CONSUMER
from mesh.prsim import Metadata


@spi(name="consumer", pattern=CONSUMER, priority=0x80000000)
class ConsumerFilter(Filter):

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        codec: Codec = ServiceLoader.load(Codec).get(MeshFlag.JSON.get_name())
        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        for key, value in Mesh.context().get_attachments().items():
            attachments[key] = value
        attachments[Metadata.MESH_TRACE_ID.key()] = Mesh.context().get_trace_id()
        attachments[Metadata.MESH_SPAN_ID.key()] = Mesh.context().get_span_id()
        attachments[Metadata.MESH_TIMESTAMP.key()] = f"{Mesh.context().get_timestamp()}"
        attachments[Metadata.MESH_RUN_MODE.key()] = f"{Mesh.context().get_run_mode().value}"
        attachments[Metadata.MESH_URN.key()] = Mesh.context().get_urn()
        attachments[Metadata.MESH_CONSUMER.key()] = codec.encode_string(Mesh.context().get_consumer())
        attachments[Metadata.MESH_PROVIDER.key()] = codec.encode_string(Mesh.context().get_provider())

        digest = Digest()
        try:
            ret = invoker.run(invocation)
            digest.write("C", MeshCode.SUCCESS.get_code())
            return ret
        except BaseException as e:
            if isinstance(e, Codeable):
                digest.write("C", e.get_code())
            else:
                digest.write("C", MeshCode.SYSTEM_ERROR.get_code())
            if isinstance(e, MeshException):
                log.error(f"{digest.trace_id},{Mesh.context().get_urn()},{e.get_message()}")
            raise e



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import mesh.log as log
import typing
import unittest

from mesh.macro import ServiceLoader
from mesh.mpc import Compiler


class TestCompile(unittest.TestCase):

    def test_compile_code(self):
        compiler = ServiceLoader.load(Compiler).get_default()
        proxy = compiler.compile("""
from mesh.macro import index

class Terminator:

    baz: str

    def get_name(self)->str:
        return "I am back!"
    
    @index(0)
    def get_index(self)->int:
        return 9
""", "proxy", 1)
        terminator = proxy.Terminator()
        name = terminator.get_name()
        log.info(name)
        members = typing.get_type_hints(proxy.Terminator)
        log.info(members)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


from abc import ABC, abstractmethod
from typing import Dict, Generic, List

from mesh.kinds import Principal
from mesh.macro import T, ServiceLoader, spi


@spi("mesh")
class Routable(ABC, Generic[T]):

    @abstractmethod
    def __init__(self, reference: T = None, attachments: Dict[str, str] = None, address: str = ""):
        pass

    def __call__(self, *args, **kwargs):
        return self

    @abstractmethod
    def within(self, key: str, value: str) -> "Routable[T]":
        """
        Route with attachments.
        :param key: attachment key
        :param value: attachment value
        :return:
        """
        pass

    @abstractmethod
    def with_map(self, attachments: Dict[str, str]) -> "Routable[T]":
        """
        Invoke the service in local network.
        :param attachments: attachments
        :return:
        """
        pass

    @abstractmethod
    def with_address(self, address: str) -> "Routable[T]":
        """
        Invoke the service in many network, it may be local or others. Broadcast mode.
        :param address: Network address.
        :return: Service invoker.
        """
        pass

    @abstractmethod
    def local(self) -> T:
        """
        Invoke the service in local network.
        :return:
        """
        pass

    @abstractmethod
    def any(self, principal: Principal) -> T:
        """
        Invoke the service in a network, it may be local or others.
        :param principal: Network principal.
        :return:Service invoker.
        """
        pass

    @abstractmethod
    def any_inst(self, inst_id: str) -> T:
        """
        Invoke the service in a network, it may be local or others.
        :param inst_id: Network principal of inst_id.
        :return:Service invoker.
        """
        pass

    @abstractmethod
    def many(self, principals: List[Principal]) -> List[T]:
        """
        Invoke the service in many network, it may be local or others. Broadcast mode.
        :param principals: Network principals.
        :return: Service invoker.
        """
        pass

    @abstractmethod
    def many_inst(self, inst_ids: List[str]) -> List[T]:
        """
        Invoke the service in many network, it may be local or others. Broadcast mode.
        :param inst_ids: Network principals.
        :return: Service invoker.
        """
        pass

    @staticmethod
    def of(reference: T) -> "Routable[T]":
        """
        Wrap a service with streamable ability.
        :param reference:  Service reference.
        :return: Streamable program interface.
        """
        cls = ServiceLoader.load(Routable).get_default_cls()
        return cls(reference, {}, "")



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any, List

from mesh.kinds.entity import Entity
from mesh.kinds.principal import Principal
from mesh.macro import ServiceLoader
from mesh.macro import index, serializable, Binding


@serializable
class Topic:

    @index(0)
    def topic(self) -> str:
        return ''

    @index(5)
    def code(self) -> str:
        return ''

    @index(10)
    def group(self) -> str:
        return ''

    @index(15)
    def sets(self) -> str:
        return ''

    def matches(self, bindings: List[Binding]) -> bool:
        if not bindings:
            return False
        for b in bindings:
            if b.topic == self.topic and b.code == self.code:
                return True
        return False


class Event:
    """Any fixed information of principal."""

    @index(0)
    def version(self) -> str:
        return ""

    @index(5)
    def tid(self) -> str:
        return ""

    @index(10)
    def sid(self) -> str:
        return ""

    @index(15)
    def eid(self) -> str:
        return ""

    @index(20)
    def mid(self) -> str:
        return ""

    @index(25)
    def timestamp(self) -> str:
        return ""

    @index(30)
    def source(self) -> Principal:
        return Principal()

    @index(35)
    def target(self) -> Principal:
        return Principal()

    @index(40)
    def binding(self) -> Topic:
        return Topic()

    @index(45)
    def entity(self) -> Entity:
        return Entity()

    @staticmethod
    def new_instance(payload: Any, topic: Topic) -> "Event":
        """
        Create local event instance.
        :param payload:
        :param topic:
        :return:
        """
        from mesh.prsim.network import Network
        network = ServiceLoader.load(Network).get_default()
        target = Principal()
        target.node_id = network.get_environ().node_id
        target.inst_id = network.get_environ().inst_id
        return Event.new_instance_with_target(payload, topic, target)

    @staticmethod
    def new_instance_with_target(payload: Any, topic: Topic, target: Principal) -> "Event":
        """
        Create any node event instance.
        :param payload:
        :param topic:
        :param target:
        :return:
        """
        from mesh.prsim.network import Network
        network = ServiceLoader.load(Network).get_default()
        source = Principal()
        source.node_id = network.get_environ().node_id
        source.inst_id = network.get_environ().inst_id
        return Event.new_instance_with_target_source(payload, topic, target, source)

    @staticmethod
    def new_instance_with_target_source(payload: Any, topic: Topic, target: Principal, source: Principal) -> "Event":
        """
        Create an event with source node and target node.
        :param payload:
        :param topic:
        :param target:
        :param source:
        :return:
        """
        event = Event()
        event.version = "1.0.0"
        event.tid = ""
        event.sid = ""
        event.eid = ""
        event.mid = ""
        event.timestamp = ""
        event.source = source
        event.target = target
        event.binding = topic
        event.entity = Entity.wrap(payload)
        return event



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import List

from mesh.kinds import Registration
from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import Registry


@spi("mesh")
class MeshRegistry(Registry):

    def __init__(self):
        self.proxy = ServiceProxy.default_proxy(Registry)

    def register(self, registration: Registration):
        return self.proxy.register(registration)

    def registers(self, registrations: List[Registration]):
        return self.proxy.registers(registrations)

    def unregister(self, registration: Registration):
        return self.proxy.unregister(registration)

    def export(self, kind: str) -> List[Registration]:
        return self.proxy.export()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import unittest

import mesh.asm as asm
import mesh.log as log
from mesh.macro import mpi
from mesh.prsim import Network


class TestGrpc(unittest.TestCase):

    @mpi
    def network(self) -> Network:
        """"""
        pass

    def test_get_environ(self):
        asm.init()
        environ = self.network.get_environ()
        assert environ
        log.info(environ.node_id)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import Type, Dict, Any

from mesh.macro import spi, T


@spi("mesh")
class Dispatcher(ABC):

    @abstractmethod
    def reference(self, mpi: Type[T]) -> T:
        """
        Refer an generic dispatcher.
        :param mpi:
        :return:
        """
        pass

    @abstractmethod
    def invoke(self, urn: str, param: Dict[str, Any]) -> Any:
        """
        Dispatch generic invoke with urn.
        :param urn:
        :param param:
        :return:
        """
        pass

    @abstractmethod
    def invoke_generic(self, urn: str, param: Any) -> Any:
        """
        Dispatch generic invoke with urn and generic params.
        :param urn:
        :param param:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import abstractmethod, ABC

from mesh.macro import spi


@spi("mesh")
class RuntimeHook(ABC):

    @abstractmethod
    def start(self):
        """
         Trigger when mesh runtime is start.
        :return:
        """
        pass

    @abstractmethod
    def stop(self):
        """
        Trigger when mesh runtime is stop.
        :return:
        """
        pass

    @abstractmethod
    def refresh(self):
        """
        Trigger then mesh runtime context is refresh or metadata is refresh.
        :return:
        """
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import abstractmethod, ABC
from typing import Type, Any, List

from mesh.kinds import Service, Reference
from mesh.macro import mpi, T, spi, Inspector
from mesh.mpc.invoker import Execution


@spi("mesh")
class Eden(ABC):

    @abstractmethod
    def define(self, metadata: mpi, reference: Type[T]) -> T:
        """
        Define the reference object.

        :param reference:
        :param metadata: Object reference type.
        :param c: Meta custom annotation.
        :return: Reference proxy
        """
        pass

    @abstractmethod
    def refer(self, metadata: mpi, reference: Type[T], method: Inspector) -> Execution[Reference]:
        """
        Refer the service reference by method.

        :param metadata:
        :param reference:
        :param method:
        :return:
        """
        pass

    @abstractmethod
    def store(self, kind: Type[T], service: Any):
        """
        Store the service object.

        :param kind:
        :param service:
        :return:
        """
        pass

    @abstractmethod
    def infer(self, urn: str) -> Execution[Service]:
        """
        Infer the reference service by domain.

        :param urn:
        :return:
        """
        pass

    @abstractmethod
    def refer_types(self) -> List[Type[T]]:
        """
        Get all reference types.
        :return: All reference types.
        """
        pass

    @abstractmethod
    def infer_types(self) -> List[Type[T]]:
        """
        Get all service types.
        :return: All service types.
        """
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import mesh.mpc.consumer_filter
import mesh.mpc.provider_filter
import mesh.mpc.robust_filter
from mesh.context import Mesh
from mesh.context import MeshContext
from mesh.mpc.compiler import Compiler
from mesh.mpc.consumer import Consumer
from mesh.mpc.eden import Eden
from mesh.mpc.factory import Factory
from mesh.mpc.filter import Filter
from mesh.mpc.generic import GenericExecution
from mesh.mpc.invoker import Invoker, Invocation, Execution
from mesh.mpc.mesh_eden import *
from mesh.mpc.provider import Provider
from mesh.mpc.schema import Schema
from mesh.mpc.service_classloader import ServiceClassLoader
from mesh.mpc.service_proxy import ServiceProxy
from mesh.mpc.stream import MeshRoutable
from mesh.mpc.transporter import Transporter, PROVIDER, CONSUMER
from mesh.mpc.urn import MESH_DOMAIN, URN, URNFlag

__all__ = (
    "Mesh",
    "URN",
    "URNFlag",
    "Consumer",
    "Eden",
    "Execution",
    "Factory",
    "Filter",
    "Invocation",
    "Invoker",
    "Provider",
    "Schema",
    "ServiceClassLoader",
    "ServiceProxy",
    "GenericExecution",
    "Compiler",
    "MESH_DOMAIN",
    "MeshRoutable",
    "Transporter",
    "PROVIDER",
    "CONSUMER",
    "MeshContext",
)


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import unittest

from mesh.codec import Codec
from mesh.macro import ServiceLoader


class TestServiceLoad(unittest.TestCase):

    def test_load(self):
        codec = ServiceLoader.load(Codec).load()


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
#
# import glob
# from os.path import join, dirname, basename, isfileø
# modules = glob.glob(join(dirname(__file__), "*.py"))
# __all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]
from mesh.kinds.commerce import CommerceLicense, CommerceEnviron
from mesh.kinds.document import Document
from mesh.kinds.entity import CacheEntity, Entity
from mesh.kinds.environ import Environ
from mesh.kinds.event import Event, Topic
from mesh.kinds.inbound import Inbound
from mesh.kinds.institution import Institution
from mesh.kinds.license import License
from mesh.kinds.location import Location
from mesh.kinds.meshflag import MeshFlag
from mesh.kinds.outbound import Outbound
from mesh.kinds.page import Page
from mesh.kinds.paging import Paging
from mesh.kinds.principal import Principal
from mesh.kinds.profile import Profile
from mesh.kinds.reference import Reference
from mesh.kinds.registration import Registration, Resource, Binding, MPS, Forward, Proxy
from mesh.kinds.route import Route
from mesh.kinds.script import Script
from mesh.kinds.service import Service
from mesh.kinds.timeout import Timeout
from mesh.kinds.versions import Versions

__all__ = (
    "Environ",
    "Route",
    "Location",
    "Principal",
    "Inbound",
    "Institution",
    "Outbound",
    "Reference",
    "Service",
    "Event",
    "Topic",
    "Entity",
    "MeshFlag",
    "CacheEntity",
    "Timeout",
    "Registration",
    "Versions",
    "Profile",
    "License",
    "CommerceLicense",
    "CommerceEnviron",
    "Document",
    "Paging",
    "Page",
    "Resource",
    "Binding",
    "MPS",
    "Forward",
    "Proxy",
    "Script",
)


def init():
    """ init function """
    pass
